import os
import sys
import h5py
import numpy as np
from skimage.io import imread

code_dir = "/home/user/BiaPy"
map_code_dir = "/home/user/mAP_3Dvolume"
input_dir = "/home/user/input"
gt_dir = "/home/user/input_gt"
temporal_dir= "/home/user/TEMP"

iou = True
mAP = True
matching_stats = True
matching_stats_ths = [0.3, 0.5, 0.75]
matching_segcompare = True
verbose = True


gt_partial_files_dir = os.path.join(temporal_dir, "GT")
os.makedirs(gt_partial_files_dir, exist_ok=True)

sys.path.insert(0, code_dir)
from biapy.utils.matching import matching, match_using_segCompare
from biapy.utils.util import save_tif_pair_discard, wrapper_matching_dataset_lazy, wrapper_matching_segCompare
from biapy.engine.metrics import jaccard_index_numpy, voc_calculation

if iou:
    all_iou = 0
    all_ov_iou = 0  # (overall IoU)
if mAP:
    # Prepare mAP call
    sys.path.insert(0, map_code_dir)
    from demo_modified import main as mAP_calculation
    class Namespace:
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

    mAP_50_total = 0
    mAP_75_total = 0
if matching_stats:
    all_matching_stats = []
if matching_segcompare:
    all_matching_stats_segCompare = []

ids = sorted(next(os.walk(input_dir))[2])
for n, id_ in enumerate(ids):
    img = imread(os.path.join(input_dir, id_)).astype(np.int64)
    mask = imread(os.path.join(gt_dir, id_)).astype(np.int64)

    print(" ")
    print("#######################################")
    print("Analizing file {}".format(os.path.join(input_dir, id_)))
    file_dir = os.path.join(temporal_dir, id_)
    os.makedirs(file_dir, exist_ok=True)

    ###########################
    # IoU & VOC (overall IoU) #
    ###########################
    if iou:
        _iou = jaccard_index_numpy((mask>0.5).astype(np.uint8), (img>0.5).astype(np.uint8))
        all_iou += _iou
        _ov_iou = voc_calculation((mask>0.5).astype(np.uint8), (img>0.5).astype(np.uint8), _iou)
        all_ov_iou += _ov_iou
        if verbose: print("Foreground IoU: {} - Overall IoU {}".format(_iou,_ov_iou))

    #######
    # mAP #
    #######
    if mAP:
        # Convert the prediction into an .h5 file
        h5file_name = os.path.join(file_dir, os.path.splitext(id_)[0]+'.h5')
        if verbose: print("Creating prediction h5 file to calculate mAP: {}".format(h5file_name))
        h5f = h5py.File(h5file_name, 'w')
        h5f.create_dataset('dataset', data=img, compression="lzf")
        h5f.close()

        # Create GT H5 file if it does not exist
        gt_f = os.path.join(gt_partial_files_dir, os.path.splitext(id_)[0]+'.h5')
        if not os.path.isfile(gt_f):
            test_file = os.path.join(gt_dir, id_)
            if verbose: print("GT .h5 file needed for mAP calculation is not found in {} so it will be created "
                              "from its mask: {}".format(gt_f, test_file))

            if not os.path.isfile(test_file):
                raise ValueError("The mask is supossed to have the same name as the image")

            if verbose: print("Saving .h5 GT data from array shape: {}".format(mask.squeeze().shape))
            os.makedirs(gt_partial_files_dir, exist_ok=True)
            h5f = h5py.File(gt_f, 'w')
            h5f.create_dataset('dataset', data=mask.squeeze(), compression="lzf")
            h5f.close()

        # Calculate mAP
        args = Namespace(gt_seg=gt_f, predict_seg=h5file_name, predict_score='', threshold="5e3, 3e4", threshold_crumb=64,
                         chunk_size=250, output_name=file_dir, do_txt=1, do_eval=1, slices="-1")
        mAP_calculation(args)

        with open(os.path.join(temporal_dir, id_, 'nucmm_map.txt'), "r") as read_obj:
            for line in read_obj:
                if 'Average Precision  (AP) @[ IoU=0.50      | area=   all | maxDets=100 ] =' in line:
                    mAP_50_total += float(line.split()[-1])
                if 'Average Precision  (AP) @[ IoU=0.75      | area=   all | maxDets=100 ] =' in line:
                    mAP_75_total += float(line.split()[-1])

    ##################
    # matching stats #
    ##################
    if matching_stats:
        r_stats = matching(mask, img, thresh=matching_stats_ths, report_matches=False)
        if verbose: print(r_stats)
        all_matching_stats.append(r_stats)

    #######################
    # matching segcompare #
    #######################
    if matching_segcompare:
        r_stats_segCompare = match_using_segCompare(mask, img)
        if verbose: print(r_stats_segCompare)
        all_matching_stats_segCompare.append(r_stats_segCompare)

print("#################")
print("# FINAL RESULTS #")
print("#################")
print("")

if iou:
    all_iou = all_iou / len(ids)
    all_ov_iou = all_ov_iou / len(ids)
    print("~~~~~~ Foreground IoU & Overall IoU ~~~~~~")
    print("Foreground IoU: {}".format(all_iou))
    print("Overall IoU: {}".format(all_ov_iou))
    print("")

if mAP:
    mAP_50_total = mAP_50_total / len(ids)
    mAP_75_total = mAP_75_total / len(ids)
    print("~~~~~~ mAP ~~~~~~")
    print("Average Precision (AP) - IoU=0.50 : {}".format(mAP_50_total))
    print("Average Precision (AP) - IoU=0.75 : {}".format(mAP_75_total))
    print("")

if matching_stats:
    stats = wrapper_matching_dataset_lazy(all_matching_stats, matching_stats_ths)
    print("~~~~~~ Matching stats ~~~~~~")
    for i in range(len(matching_stats_ths)):
        print("IoU TH={}".format(matching_stats_ths[i]))
        print(stats[i])
    print("")

if matching_segcompare:
    print("~~~~~~ Matching segcompare ~~~~~~")
    stats_segCompare = wrapper_matching_segCompare(all_matching_stats_segCompare)
    print("segCompare segmentation rates:")
    print(stats_segCompare)
    print("")

print("Remember to remove the temporal folder: {}".format(temporal_dir))



import math
import torch
import numpy as np
from tqdm import tqdm
from torchmetrics.image import PeakSignalNoiseRatio

from biapy.data.data_2D_manipulation import crop_data_with_overlap, merge_data_with_overlap
from biapy.data.data_3D_manipulation import crop_3D_data_with_overlap, merge_3D_data_with_overlap
from biapy.data.post_processing.post_processing import ensemble8_2d_predictions, ensemble16_3d_predictions
from biapy.utils.util import save_tif
from biapy.utils.misc import to_pytorch_format, to_numpy_format
from biapy.engine.base_workflow import Base_Workflow
from biapy.engine.metrics import dfcan_loss
from biapy.data.pre_processing import normalize, denormalize, undo_norm_range01

class Super_resolution_Workflow(Base_Workflow):
    """
    Semantic segmentation workflow where the goal is to assign a class to each pixel of the input image. 
    More details in `our documentation <https://biapy.readthedocs.io/en/latest/workflows/super_resolution.html>`_.  

    Parameters
    ----------
    cfg : YACS configuration
        Running configuration.
    
    Job_identifier : str
        Complete name of the running job.

    device : Torch device
        Device used. 

    args : argpase class
        Arguments used in BiaPy's call. 
    """
    def __init__(self, cfg, job_identifier, device, args, **kwargs):
        super(Super_resolution_Workflow, self).__init__(cfg, job_identifier, device, args, **kwargs)
        self.stats['psnr_merge_patches'] = 0

        # From now on, no modification of the cfg will be allowed
        self.cfg.freeze()

        # Activations for each output channel:
        # channel number : 'activation'
        self.activations = [{':': 'Linear'}]

        # Workflow specific training variables
        self.mask_path = cfg.DATA.TRAIN.GT_PATH
        self.load_Y_val = True

    def define_metrics(self):
        """
        Definition of self.metrics, self.metric_names and self.loss variables.
        """
        self.metrics = [PeakSignalNoiseRatio()]
        self.metric_names = ["PSNR"]
        if self.cfg.MODEL.ARCHITECTURE == 'dfcan':
            print("Overriding 'LOSS.TYPE' to set it to DFCAN loss")
            self.loss = dfcan_loss(self.device)
        else:
            print("Overriding 'LOSS.TYPE' to set it to MAE")
            self.loss = torch.nn.L1Loss()

    def metric_calculation(self, output, targets, metric_logger=None):
        """
        Execution of the metrics defined in :func:`~define_metrics` function. 

        Parameters
        ----------
        output : Torch Tensor
            Prediction of the model. 

        targets : Torch Tensor
            Ground truth to compare the prediction with. 

        metric_logger : MetricLogger, optional
            Class to be updated with the new metric(s) value(s) calculated. 
        
        Returns
        -------
        value : float
            Value of the metric for the given prediction. 
        """
        # Denormalization to calculate PSNR with original range values 
        output = output.to(torch.float32).detach().cpu()
        targets = targets.to(torch.float32).detach().cpu()
        if self.data_norm['type'] == 'div':
            if len([x for x in list(self.data_norm.keys()) if not 'reduced' in x]) > 0:
                output = torch.clip(output*255, 0, 255) 
                targets = torch.round(targets*255) 
            else:
                output = torch.clip(output*65535, 0, 65535)
                targets = torch.round(targets*65535) 
            output = torch.round(output)
        else:
            output = (output * self.data_norm['std']) + self.data_norm['mean']
            output = torch.round(output)                                                                 
            output = output+abs(torch.min(output))

            targets = (targets * self.data_norm['std']) + self.data_norm['mean']
            targets = torch.round(targets)                                                                 
            targets = targets+abs(torch.min(targets))

        with torch.no_grad():
            train_psnr = self.metrics[0](output, targets)
            train_psnr = train_psnr.item() if not torch.isnan(train_psnr) else 0
            if metric_logger is not None:
                metric_logger.meters[self.metric_names[0]].update(train_psnr)
            else:
                return train_psnr

    def process_sample(self, norm): 
        """
        Function to process a sample in the inference phase. 

        Parameters
        ----------
        norm : List of dicts
            Normalization used during training. Required to denormalize the predictions of the model.
        """
        if self.cfg.PROBLEM.NDIM == '2D':
            original_data_shape = (self._X.shape[0], self._X.shape[1]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING,
                                   self._X.shape[2]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING, self._X.shape[3])
        else:
            original_data_shape = (self._X.shape[0], self._X.shape[1],
                                   self._X.shape[2]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING,
                                   self._X.shape[3]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING, self._X.shape[4])

        # Crop if necessary
        if self._X.shape[1:-1] != self.cfg.DATA.PATCH_SIZE[:-1]:
            if self.cfg.PROBLEM.NDIM == '2D':
                self._X = crop_data_with_overlap(self._X, self.cfg.DATA.PATCH_SIZE, overlap=self.cfg.DATA.TEST.OVERLAP, 
                    padding=self.cfg.DATA.TEST.PADDING, verbose=self.cfg.TEST.VERBOSE)
            else:
                self._X = crop_3D_data_with_overlap(self._X[0], self.cfg.DATA.PATCH_SIZE, overlap=self.cfg.DATA.TEST.OVERLAP, 
                    padding=self.cfg.DATA.TEST.PADDING, verbose=self.cfg.TEST.VERBOSE)

        # Predict each patch
        pred = []
        if self.cfg.TEST.AUGMENTATION:
            for k in tqdm(range(self._X.shape[0]), leave=False):
                if self.cfg.PROBLEM.NDIM == '2D':
                    p = ensemble8_2d_predictions(self._X[k], n_classes=self.cfg.MODEL.N_CLASSES,
                            pred_func=(
                                lambda img_batch_subdiv: 
                                    to_numpy_format(
                                        self.apply_model_activations(
                                            self.model(to_pytorch_format(img_batch_subdiv, self.axis_order, self.device)),
                                            ), 
                                        self.axis_order_back
                                    )
                            )
                        )
                else:
                    p = ensemble16_3d_predictions(self._X[k], batch_size_value=self.cfg.TRAIN.BATCH_SIZE,
                            pred_func=(
                                lambda img_batch_subdiv: 
                                    to_numpy_format(
                                        self.apply_model_activations(
                                            self.model(to_pytorch_format(img_batch_subdiv, self.axis_order, self.device)),
                                            ), 
                                        self.axis_order_back
                                    )
                            )
                        )
                pred.append(np.expand_dims(p,0))
        else:
            self._X = to_pytorch_format(self._X, self.axis_order, self.device)
            l = int(math.ceil(self._X.shape[0]/self.cfg.TRAIN.BATCH_SIZE))
            for k in tqdm(range(l), leave=False):
                top = (k+1)*self.cfg.TRAIN.BATCH_SIZE if (k+1)*self.cfg.TRAIN.BATCH_SIZE < self._X.shape[0] else self._X.shape[0]
                with torch.cuda.amp.autocast():
                    p = self.model(self._X[k*self.cfg.TRAIN.BATCH_SIZE:top])
                    p = to_numpy_format(self.apply_model_activations(p), self.axis_order_back)
                pred.append(p)
        del self._X, p

        # Reconstruct the predictions
        pred = np.concatenate(pred)
        if original_data_shape[1:-1] != self.cfg.DATA.PATCH_SIZE[:-1]:
            if self.cfg.PROBLEM.NDIM == '3D': original_data_shape = original_data_shape[1:]
            f_name = merge_data_with_overlap if self.cfg.PROBLEM.NDIM == '2D' else merge_3D_data_with_overlap
            if self.cfg.PROBLEM.NDIM == '2D':
                pad = tuple(p*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING for p in self.cfg.DATA.TEST.PADDING)
                ov = tuple(o*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING for o in self.cfg.DATA.TEST.OVERLAP)
            else:
                pad = (self.cfg.DATA.TEST.PADDING[0], 
                       self.cfg.DATA.TEST.PADDING[1]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING,
                       self.cfg.DATA.TEST.PADDING[2]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING)
                ov = (self.cfg.DATA.TEST.OVERLAP[0], 
                      self.cfg.DATA.TEST.OVERLAP[1]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING,
                      self.cfg.DATA.TEST.OVERLAP[2]*self.cfg.PROBLEM.SUPER_RESOLUTION.UPSCALING)
            pred = f_name(pred, original_data_shape[:-1]+(pred.shape[-1],), padding=pad, 
                overlap=ov, verbose=self.cfg.TEST.VERBOSE)
        else:
            pred = pred[0]

        # Undo normalization
        x_norm = norm[0]
        if x_norm['type'] == 'div':
            pred = undo_norm_range01(pred, x_norm)
        else:
            pred = denormalize(pred, x_norm['mean'], x_norm['std'])  
            
            if x_norm['orig_dtype'] not in [np.dtype('float64'), np.dtype('float32'), np.dtype('float16')]:
                pred = np.round(pred)
                minpred = np.min(pred)                                                                                                
                pred = pred+abs(minpred)

            pred = pred.astype(x_norm['orig_dtype'])

        # Save image
        if self.cfg.PATHS.RESULT_DIR.PER_IMAGE != "":
            save_tif(np.expand_dims(pred,0), self.cfg.PATHS.RESULT_DIR.PER_IMAGE, self.processing_filenames, 
                verbose=self.cfg.TEST.VERBOSE)
    
        # Calculate PSNR
        if pred.dtype == np.dtype('uint16'):
            pred = pred.astype(np.float32)
        if self._Y.dtype == np.dtype('uint16'):
            self._Y = self._Y.astype(np.float32)
        if self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST:
            psnr_merge_patches = self.metrics[0](torch.from_numpy(pred), torch.from_numpy(self._Y))
            self.stats['psnr_merge_patches'] += psnr_merge_patches

    def torchvision_model_call(self, in_img, is_train=False):
        """
        Call a regular Pytorch model.

        Parameters
        ----------
        in_img : Tensor
            Input image to pass through the model.

        is_train : bool, optional
            Whether if the call is during training or inference. 

        Returns
        -------
        prediction : Tensor 
            Image prediction. 
        """
        pass

    def after_merge_patches(self, pred):
        """
        Steps need to be done after merging all predicted patches into the original image.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        pass

    def after_merge_patches_by_chunks_proccess_patch(self, filename):
        """
        Place any code that needs to be done after merging all predicted patches into the original image
        but in the process made chunk by chunk. This function will operate patch by patch defined by 
        ``DATA.PATCH_SIZE``.

        Parameters
        ----------
        filename : List of str
            Filename of the predicted image H5/Zarr.  
        """
        pass

    def after_full_image(self, pred):
        """
        Steps that must be executed after generating the prediction by supplying the entire image to the model.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        pass

    def after_all_images(self):
        """
        Steps that must be done after predicting all images. 
        """
        pass

    def normalize_stats(self, image_counter):
        """
        Normalize statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to average the metrics.
        """
        self.stats['psnr_merge_patches'] = self.stats['psnr_merge_patches'] / image_counter

    def print_stats(self, image_counter):
        """
        Print statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to call ``normalize_stats``.
        """
        self.normalize_stats(image_counter)

        if self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST:
            print("Test PSNR (merge patches): {}".format(self.stats['psnr_merge_patches']))
            print(" ")


        


import argparse
import os
import sys

from biapy import BiaPy

if __name__ == '__main__':

    ##########################
    #   ARGS COMPROBATION    #
    ##########################

    # Normal exec:
    # python -u main.py \
    #     --config $input_job_cfg_file \        
    #     --result_dir $result_dir \        
    #     --name $job_name \          
    #     --run_id $job_counter \        
    #     --gpu 0
    # Distributed:
    # python -u -m torch.distributed.run \
    #     --nproc_per_node=2 \  
    #     main.py \
    #     --config $input_job_cfg_file \       
    #     --result_dir $result_dir \        
    #     --name $job_name \          
    #     --run_id $job_counter \        
    #     --gpu 0,1

    parser = argparse.ArgumentParser()
    parser.add_argument("--config", required=True, help="Path to the configuration file")
    parser.add_argument("--result_dir", help="Path to where the resulting output of the job will be stored",
                        default=os.getenv('HOME'))
    parser.add_argument("--name", help="Job name", default="unknown_job")
    parser.add_argument("--run_id", help="Run number of the same job", type=int, default=1)
    parser.add_argument("--gpu", help="GPU number according to 'nvidia-smi' command", type=str)

    # Distributed training parameters
    parser.add_argument('--world_size', default=1, type=int,
                        help='number of distributed processes')
    parser.add_argument('--local_rank', default=-1, type=int,
                        help='Node rank for distributed training. Necessary for using the torch.distributed.launch utility.')
    parser.add_argument('--dist_on_itp', action='store_true')
    parser.add_argument('--dist_url', default='env://',
                        help='url used to set up distributed training')
    parser.add_argument('--dist_backend', type=str, default='nccl', choices=['nccl', 'gloo'],
                        help='Backend to use in distributed mode')
    args = parser.parse_args()

    _biapy = BiaPy(**vars(args))
    _biapy.run_job()
    sys.exit(0)


import os
import sys
import numpy as np
from skimage.io import imread
from tqdm import tqdm

code_dir = "/home/user/BiaPy"
input_dir = "/home/user/datasets/x"
output_dir = "/home/user/datasets/x_out"
crop_shape = (64, 64, 64, 1)
overlap = (0, 0, 0)
padding = (0, 0, 0)
median_padding = False

sys.path.insert(0, code_dir)
from biapy.data.data_3D_manipulation import crop_3D_data_with_overlap
from biapy.utils.util import save_tif

ids = sorted(next(os.walk(input_dir))[2])
for n, id_ in tqdm(enumerate(ids), total=len(ids)):
    img = imread(os.path.join(input_dir, id_))

    if len(img.shape) == 3:
        img = np.expand_dims(img, axis=-1)

    if img.shape != crop_shape[:3]+(img.shape[-1],):
        img = crop_3D_data_with_overlap(img, crop_shape[:3]+(img.shape[-1],), overlap=overlap, padding=padding,
                                        median_padding=median_padding, verbose=True)
    else:
        img = np.expand_dims(img, axis=0)

    filenames = []
    d = len(str(len(img)))
    for k in range(img.shape[0]):
        filenames.append(os.path.splitext(id_)[0]+"_crop"+str(k).zfill(d)+'.tif')
    save_tif(img, output_dir, filenames)

print("Finished!")



import math
import torch
import torch.nn as nn
import torch.nn.init as init

class wdsr(nn.Module):
  """
  WDSR model. 
  
  Reference: `Wide Activation for Efficient and Accurate Image Super-Resolution <https://arxiv.org/abs/1808.08718>`_. 

  Adapted from `here <https://github.com/yjn870/WDSR-pytorch/tree/master>`_.
  """
  def __init__(self, scale, num_filters=32, num_res_blocks=16, res_block_expansion=6, num_channels=1):
    super(wdsr, self).__init__()
    kernel_size = 3
    skip_kernel_size = 5
    weight_norm = torch.nn.utils.weight_norm
    num_outputs = scale * scale * num_channels

    body = []
    conv = weight_norm(
        nn.Conv2d(
            num_channels,
            num_filters,
            kernel_size,
            padding=kernel_size // 2))
    init.ones_(conv.weight_g)
    init.zeros_(conv.bias)
    body.append(conv)
    for _ in range(num_res_blocks):
      body.append(
          Block(
              num_filters,
              kernel_size,
              res_block_expansion,
              weight_norm=weight_norm,
              res_scale=1 / math.sqrt(num_res_blocks),
          ))
    conv = weight_norm(
        nn.Conv2d(
            num_filters,
            num_outputs,
            kernel_size,
            padding=kernel_size // 2))
    init.ones_(conv.weight_g)
    init.zeros_(conv.bias)
    body.append(conv)
    self.body = nn.Sequential(*body)

    skip = []
    if num_channels != num_outputs:
      conv = weight_norm(
          nn.Conv2d(
            num_channels,
            num_outputs,
            skip_kernel_size,
            padding=skip_kernel_size // 2))
      init.ones_(conv.weight_g)
      init.zeros_(conv.bias)
      skip.append(conv)
    self.skip = nn.Sequential(*skip)

    shuf = []
    if scale > 1:
      shuf.append(nn.PixelShuffle(scale))
    self.shuf = nn.Sequential(*shuf)

  def forward(self, x):
    x = self.body(x) + self.skip(x)
    x = self.shuf(x)
    x = torch.clamp(x, min=0, max=1)
    return x


class Block(nn.Module):
  def __init__(self,
               num_residual_units,
               kernel_size,
               width_multiplier=1,
               weight_norm=torch.nn.utils.weight_norm,
               res_scale=1):
    super(Block, self).__init__()
    body = []
    conv = weight_norm(
        nn.Conv2d(
            num_residual_units,
            int(num_residual_units * width_multiplier),
            kernel_size,
            padding=kernel_size // 2))
    init.constant_(conv.weight_g, 2.0)
    init.zeros_(conv.bias)
    body.append(conv)
    body.append(nn.ReLU(True))
    conv = weight_norm(
        nn.Conv2d(
            int(num_residual_units * width_multiplier),
            num_residual_units,
            kernel_size,
            padding=kernel_size // 2))
    init.constant_(conv.weight_g, res_scale)
    init.zeros_(conv.bias)
    body.append(conv)

    self.body = nn.Sequential(*body)

  def forward(self, x):
    x = self.body(x) + x
    return x


import torch
import numpy as np
import torch.nn as nn
from typing import Callable, List, Optional, Tuple

class PatchEmbed(nn.Module):
    """ 2D Image to Patch Embedding
    """
    def __init__(
            self,
            img_size: int = 224,
            patch_size: int = 16,
            in_chans: int = 3,
            ndim: int = 2, 
            embed_dim: int = 768,
            norm_layer: Optional[Callable] = None,
            flatten: bool = True,
            bias: bool = True,
            strict_img_size: bool = True,
    ):
        super().__init__()
        self.ndim = ndim
        self.patch_size = patch_size
        self.img_size = img_size
        self.grid_size = self.img_size // self.patch_size
        self.num_patches = self.grid_size**self.ndim
        self.flatten = flatten
        self.strict_img_size = strict_img_size

        if self.ndim == 2:
            self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size, bias=bias)
        else:
            self.proj = nn.Conv3d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size, bias=bias)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()
        
    def forward(self, x):
        if self.ndim == 2:
            B, C, H, W = x.shape
            Z = -1
        else:
            B, C, Z, H, W = x.shape
        if self.strict_img_size:
            assert H == self.img_size, f"Input height ({H}) doesn't match model ({self.img_size})."
            assert W == self.img_size, f"Input width ({W}) doesn't match model ({self.img_size})."

        x = self.proj(x)
        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # NCHW -> NLC
        x = self.norm(x)
        return x


import cv2
import random
import math
import numpy as np
from PIL import Image
from skimage.transform import resize
from skimage.draw import line
from scipy.ndimage.measurements import label
from scipy.ndimage.morphology import binary_dilation
from scipy.ndimage import rotate


def cutout(img, mask, channels, z_size, nb_iterations=(1,3), size=(0.2,0.4), cval=0, res_relation=(1,1), apply_to_mask=False):
    """Cutout data augmentation presented in `Improved Regularization of Convolutional Neural Networks with Cutout
       <https://arxiv.org/pdf/1708.04552.pdf>`_.

       Parameters
       ----------
       img : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       mask : 3D Numpy array
           Mask to transform. E.g. ``(y, x, channels)``.

       channels : int
           Size of channel dimension. Used for 3D images as the channels have been merged with the z axis.

       z_size : int
           Size of z dimension. Used for 3D images as the z axis has been merged with the channels. Set to -1 to when
           do not want to be applied.

       nb_iterations : tuple of ints, optional
           Number of areas to fill the image with. E.g. ``(1, 3)``.

       size : tuple of floats, optional
           Range to choose the size of the areas to create.

       cval : int, optional
           Value to fill the area with.

       res_relation: tuple of ints/floats, optional
           Relation between axis resolution in ``(x,y,z)``. E.g. ``(1,1,0.27)`` for anisotropic data of 
           8umx8umx30um resolution.

       apply_to_mask : boolean, optional
           To apply cutout to the mask.

       Returns
       -------
       out : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       mask : 3D Numpy array
           Transformed mask. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``nb_iterations=(1,3)``, ``size=(0.05,0.3)``,
       ``apply_to_mask=False`` may result in:

       +----------------------------------------------+----------------------------------------------+
       | .. figure:: ../../../img/orig_cutout.png     | .. figure:: ../../../img/orig_cutout_mask.png|
       |   :width: 80%                                |   :width: 80%                                |
       |   :align: center                             |   :align: center                             |
       |                                              |                                              |
       |   Input image                                |   Corresponding mask                         |
       +----------------------------------------------+----------------------------------------------+
       | .. figure:: ../../../img/cutout.png          | .. figure:: ../../../img/cutout_mask.png     |
       |   :width: 80%                                |   :width: 80%                                |
       |   :align: center                             |   :align: center                             |
       |                                              |                                              |
       |   Augmented image                            |   Augmented mask                             |
       +----------------------------------------------+----------------------------------------------+

       The grid is painted for visualization purposes.
    """

    it = np.random.randint(nb_iterations[0], nb_iterations[1])

    out = img.copy()
    m_out = mask.copy()
    for i in range(it):
        _size = random.uniform(size[0], size[1])
        y_size = int(max(min(img.shape[0]*_size*res_relation[1],img.shape[0]),1))
        x_size = int(max(min(img.shape[1]*_size*res_relation[0],img.shape[1]),1))

        # Choose a random point
        cy = np.random.randint(0, img.shape[0]-y_size)
        cx = np.random.randint(0, img.shape[1]-x_size)
        if z_size != -1:
            _z_size = int(max(min(z_size*_size*res_relation[2],z_size),1))
            cz = np.random.randint(0, z_size-_z_size)

            out[cy:cy+y_size, cx:cx+x_size, cz*channels:(cz*channels)+(_z_size*channels)] = cval
            if apply_to_mask:
                m_out[cy:cy+y_size, cx:cx+x_size, cz*channels:(cz*channels)+(_z_size*channels)] = 0
        else:
            out[cy:cy+y_size, cx:cx+x_size] = cval
            if apply_to_mask:
                m_out[cy:cy+y_size, cx:cx+x_size] = 0

    return out, m_out


def cutblur(img, size=(0.2,0.4), down_ratio_range=(2,8), only_inside=True):
    """CutBlur data augmentation introduced in `Rethinking Data Augmentation for Image Super-resolution: A Comprehensive
       Analysis and a New Strategy <https://arxiv.org/pdf/2004.00448.pdf>`_ and adapted from
       https://github.com/clovaai/cutblur .


       Parameters
       ----------
       img : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       size : float, optional
           Size of the region to transform.

       down_ratio_range : tuple of ints, optional
           Downsampling ratio range to be applied. E.g. ``(2, 8)``.

       only_inside : bool, optional
           If ``True`` only the region inside will be modified (cut LR into HR image). If ``False`` the ``50%`` of the
           times the region inside will be modified (cut LR into HR image) and the other ``50%`` the inverse will be
           done (cut HR into LR image). See Figure 1 of the official `paper <https://arxiv.org/pdf/2004.00448.pdf>`_.

       Returns
       -------
       out : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``size=(0.2,0.4)``, ``down_ratio_range=(2,8)``,
       ``only_inside=True`` may result in:

       +--------------------------------------------+--------------------------------------------+
       | .. figure:: ../../../img/orig_cutblur.png  | .. figure:: ../../../img/cutblur.png       |
       |   :width: 80%                              |   :width: 80%                              |
       |   :align: center                           |   :align: center                           |
       |                                            |                                            |
       |   Input image                              |   Augmented image                          |
       +--------------------------------------------+--------------------------------------------+
       | .. figure:: ../../../img/orig_cutblur2.png | .. figure:: ../../../img/cutblur2.png      |
       |   :width: 80%                              |   :width: 80%                              |
       |   :align: center                           |   :align: center                           |
       |                                            |                                            |
       |   Input image                              |   Augmented image                          |
       +--------------------------------------------+--------------------------------------------+

       The grid and the red square are painted for visualization purposes.
    """

    _size = random.uniform(size[0], size[1])

    y_size = int(img.shape[0]*_size)
    x_size = int(img.shape[1]*_size)

    # Choose a random point
    cy = np.random.randint(0, img.shape[0]-(y_size))
    cx = np.random.randint(0, img.shape[1]-(x_size))

    # Choose a random downsampling ratio
    down_ratio = np.random.randint(down_ratio_range[0], down_ratio_range[1]+1)
    out_shape = np.array(img.shape[:2])//down_ratio

    if not only_inside:
        inside = True if random.uniform(0, 1) < 0.5 else False
    else:
        inside = True

    out = img.copy()

    # Apply cutblur to all channels
    for i in range(img.shape[-1]):
        if inside:
            temp = img[cy:cy+y_size, cx:cx+x_size, i].copy()
        else:
            temp = img[..., i].copy()

        downsampled = resize(temp, out_shape, order=1, mode='reflect',
                             clip=True, preserve_range=True, anti_aliasing=True)
        upsampled = resize(downsampled, temp.shape, order=0, mode='reflect',
                           clip=True, preserve_range=True, anti_aliasing=False)
        if inside:
            out[cy:cy+y_size, cx:cx+x_size, i] = upsampled
        else:
            temp[cy:cy+y_size, cx:cx+x_size] = img[cy:cy+y_size, cx:cx+x_size, i]
            out[...,i] = temp

    return out


def cutmix(im1, im2, mask1, mask2, size=(0.2,0.4)):
    """Cutmix augmentation introduced in `CutMix: Regularization Strategy to Train Strong Classifiers with Localizable
       Features <https://arxiv.org/abs/1905.04899>`_. With this augmentation a region of the image sample is filled
       with a given second image. This implementation is used for semantic segmentation so the masks of the images are
       also needed. It assumes that the images are of the same shape.

       Parameters
       ----------
       im1 : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       im2 : 3D Numpy array
           Image to paste into the region of ``im1``. E.g. ``(y, x, channels)``.

       mask1 : 3D Numpy array
           Mask to transform (belongs to ``im1``). E.g. ``(y, x, channels)``.

       mask2 : 3D Numpy array
           Mask to paste into the region of ``mask1``. E.g. ``(y, x, channels)``.

       size : tuple of floats, optional
           Range to choose the size of the areas to transform. E.g. ``(0.2, 0.4)``.

       Returns
       -------
       out : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       m_out : 3D Numpy array
           Transformed mask. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``size=(0.2,0.4)`` may result in:

       +----------------------------------------------+----------------------------------------------+
       | .. figure:: ../../../img/orig_cutmix.png     | .. figure:: ../../../img/orig_cutmix_mask.png|
       |   :width: 80%                                |   :width: 80%                                |
       |   :align: center                             |   :align: center                             |
       |                                              |                                              |
       |   Input image                                |   Corresponding mask                         |
       +----------------------------------------------+----------------------------------------------+
       | .. figure:: ../../../img/cutmix.png          | .. figure:: ../../../img/cutmix_mask.png     |
       |   :width: 80%                                |   :width: 80%                                |
       |   :align: center                             |   :align: center                             |
       |                                              |                                              |
       |   Augmented image                            |   Augmented mask                             |
       +----------------------------------------------+----------------------------------------------+

       The grid is painted for visualization purposes.
    """

    _size = random.uniform(size[0], size[1])

    y_size = int(im1.shape[0]*_size)
    x_size = int(im1.shape[1]*_size)

    # Choose a random point from the image to be transformed
    im1cy = np.random.randint(0, im1.shape[0]-(y_size))
    im1cx = np.random.randint(0, im1.shape[1]-(x_size))
    # Choose a random point from the other image
    im2cy = np.random.randint(0, im2.shape[0]-(y_size))
    im2cx = np.random.randint(0, im2.shape[1]-(x_size))

    out = im1.copy()
    m_out = mask1.copy()

    # Apply cutblur to all channels
    for i in range(im1.shape[-1]):
        out[im1cy:im1cy+y_size, im1cx:im1cx+x_size, i] = im2[im2cy:im2cy+y_size, im2cx:im2cx+x_size, i]
        m_out[im1cy:im1cy+y_size, im1cx:im1cx+x_size, i] = mask2[im2cy:im2cy+y_size, im2cx:im2cx+x_size, i]

    return out, m_out


def cutnoise(img, scale=(0.1,0.2), nb_iterations=(1,3), size=(0.2,0.4)):
    """Cutnoise data augmentation. Randomly add noise to a cuboid region in the image to force the model to learn
       denoising when making predictions.

       Parameters
       ----------
       img : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       scale : tuple of floats, optional
           Scale of the random noise. E.g. ``(0.1, 0.2)``.

       nb_iterations : tuple of ints, optional
           Number of areas with noise to create. E.g. ``(1, 3)``.

       size : boolean, optional
           Range to choose the size of the areas to transform. E.g. ``(0.2, 0.4)``.

       Returns
       -------
       out : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``scale=(0.1,0.2)``, ``nb_iterations=(1,3)`` and
       ``size=(0.2,0.4)`` may result in:

       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_cutnoise.png  | .. figure:: ../../../img/cutnoise.png       |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_cutnoise2.png | .. figure:: ../../../img/cutnoise2.png      |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+

       The grid and the red squares are painted for visualization purposes.
    """

    it = np.random.randint(nb_iterations[0], nb_iterations[1])

    out = img.copy()
    for i in range(it):
        _size = random.uniform(size[0], size[1])
        y_size = int(img.shape[0]*_size)
        x_size = int(img.shape[1]*_size)

        # Choose a random point
        cy = np.random.randint(0, img.shape[0]-(y_size))
        cx = np.random.randint(0, img.shape[1]-(x_size))

        max_value = np.max(out)
        _scale = random.uniform(scale[0], scale[1])*max_value
        noise = np.random.normal(loc=0, scale=_scale, size=(y_size, x_size))
        out[cy:cy+y_size, cx:cx+x_size, :] += np.stack((noise,)*out.shape[-1], axis=-1)
    return np.clip(out, 0, max_value)


def misalignment(img, mask, displacement=16, rotate_ratio=0.0, c_relation="1_1"):
    """Mis-alignment data augmentation of image stacks. This augmentation is applied to both images and masks.

       Implementation based on `PyTorch Connectomics' misalign.py
       <https://github.com/zudi-lin/pytorch_connectomics/blob/master/connectomics/data/augmentation/misalign.py>`_.

       Parameters
       ----------
       img : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       mask : 3D Numpy array
           Mask to transform. E.g. ``(y, x, channels)``

       displacement : int, optional
           Maximum pixel displacement in ``xy``-plane.

       rotate_ratio : float, optional
           Ratio of rotation-based mis-alignment.

       Returns
       -------
       out : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       m_out : 3D Numpy array
           Transformed mask. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``displacement=16`` and ``rotate_ratio=0.5`` may result in:

       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_miss.png      | .. figure:: ../../../img/orig_miss_mask.png |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Corresponding mask                        |
       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/miss.png           | .. figure:: ../../../img/miss_mask.png      |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Augmented image                           |   Augmented mask                            |
       +---------------------------------------------+---------------------------------------------+

       The grid is painted for visualization purposes.
    """

    out = np.zeros(img.shape, img.dtype)
    m_out = np.zeros(mask.shape, mask.dtype)

    # 2D
    if img.shape[-1] < 5:
        oy = np.random.randint(1, img.shape[0]-1)
        d = np.random.randint(0, displacement)
        if random.uniform(0, 1) < rotate_ratio:
            # Apply misalignment to all channels
            for i in range(img.shape[-1]):
                out[:oy,:,i] = img[:oy,:,i]
                out[oy:,:img.shape[1]-d,i] = img[oy:,d:,i]
                m_out[:oy,:,i] = mask[:oy,:,i]
                m_out[oy:,:mask.shape[1]-d,i] = mask[oy:,d:,i]
        else:
            H, W = img.shape[:2]
            M = random_rotate_matrix(H, displacement)
            H = H - oy
            # Apply misalignment to all channels
            for i in range(img.shape[-1]):
                out[:oy,:,i] = img[:oy,:,i]
                out[oy:,:,i] = cv2.warpAffine(img[oy:,:,i], M, (W,H), 1.0, flags=cv2.INTER_LINEAR,
                                              borderMode=cv2.BORDER_CONSTANT)
                m_out[:oy,:,i] = mask[:oy,:,i]
                m_out[oy:,:,i] = cv2.warpAffine(mask[oy:,:,i], M, (W,H), 1.0, flags=cv2.INTER_NEAREST,
                                                borderMode=cv2.BORDER_CONSTANT)
    # 3D
    else:
        out_shape = (img.shape[0]-displacement, img.shape[1]-displacement,
                     img.shape[2])

        mode = 'slip' if random.uniform(0, 1) < 0.5 else 'translation'

        # Calculate the amount of image and mask channels to determine which slice apply the tranformation to
        idx = np.random.randint(1, img.shape[-1]-1)
        relation = c_relation.split('_')
        img_rel = int(relation[0])
        mask_rel = int(relation[1])
        idx = int(idx - (idx%img_rel))
        idx_mask = int((idx/img_rel)*mask_rel)

        if random.uniform(0, 1) < rotate_ratio:
            out = img.copy()
            m_out = mask.copy()

            H, W = img.shape[:2]
            M = random_rotate_matrix(H, displacement)
            if mode == 'slip':
                # Apply the change to all images/masks in the last dimension that represent the slice selected. This
                # needs to be done because the last dimension is z axis mutiplied by the channels
                for k in range(img_rel):
                    out[...,idx+k] = 0
                    out[...,idx+k] = cv2.warpAffine(img[...,idx+k], M, (H,W), 1.0, flags=cv2.INTER_LINEAR,
                                                    borderMode=cv2.BORDER_CONSTANT)
                for k in range(mask_rel):
                    m_out[...,idx_mask+k] = 0
                    m_out[...,idx_mask+k] = cv2.warpAffine(mask[...,idx_mask+k], M, (H,W), 1.0, flags=cv2.INTER_NEAREST,
                                                           borderMode=cv2.BORDER_CONSTANT)
            else:
                for i in range(idx, img.shape[-1]):
                    out[...,i] = 0
                    out[...,i] = cv2.warpAffine(img[...,i], M, (H,W), 1.0, flags=cv2.INTER_LINEAR,
                                                borderMode=cv2.BORDER_CONSTANT)
                for i in range(idx_mask, mask.shape[-1]):
                    m_out[...,i] = 0
                    m_out[...,i] = cv2.warpAffine(mask[...,i], M, (H,W), 1.0, flags=cv2.INTER_NEAREST,
                                                  borderMode=cv2.BORDER_CONSTANT)
        else:
            random_state = np.random.RandomState()
            x0 = random_state.randint(displacement)
            y0 = random_state.randint(displacement)
            x1 = random_state.randint(displacement)
            y1 = random_state.randint(displacement)
            if mode == 'slip':
                out[y0:y0+out_shape[0],x0:x0+out_shape[1],:] = img[y0:y0+out_shape[0],x0:x0+out_shape[1],:]
                for k in range(img_rel):
                    out[...,idx+k] = 0
                    out[y1:y1+out_shape[0],x1:x1+out_shape[1],idx+k] = img[y1:y1+out_shape[0], x1:x1+out_shape[1],idx+k]

                m_out[y0:y0+out_shape[0], x0:x0+out_shape[1],:] = mask[y0:y0+out_shape[0], x0:x0+out_shape[1],:]
                for k in range(mask_rel):
                    m_out[...,idx_mask+k] = 0
                    m_out[y1:y1+out_shape[0],x1:x1+out_shape[1],idx_mask+k] = mask[y1:y1+out_shape[0],x1:x1+out_shape[1],idx_mask+k]
            else:
                out[y0:y0+out_shape[0],x0:x0+out_shape[1],:idx] = img[y0:y0+out_shape[0],x0:x0+out_shape[1],:idx]
                out[y1:y1+out_shape[0],x1:x1+out_shape[1],idx:] = img[y1:y1+out_shape[0],x1:x1+out_shape[1],idx:]
                m_out[y0:y0+out_shape[0],x0:x0+out_shape[1],:idx_mask] = mask[y0:y0+out_shape[0],x0:x0+out_shape[1],:idx_mask]
                m_out[y1:y1+out_shape[0],x1:x1+out_shape[1],idx_mask:] = mask[y1:y1+out_shape[0],x1:x1+out_shape[1],idx_mask:]

    return out, m_out


def random_rotate_matrix(height, displacement):
    """Auxiliary function for missaligmnet. """
    x = (displacement / 2.0)
    y = ((height - displacement) / 2.0) * 1.42
    angle = math.asin(x/y) * 2.0 * 57.2958 # convert radians to degrees
    rand_angle = (random.uniform(0, 1) - 0.5) * 2.0 * angle
    M = cv2.getRotationMatrix2D((height/2, height/2), rand_angle, 1)
    return M

def brightness(image, brightness_factor=(0,0),  mode='2D'):
    """Randomly adjust brightness between a range.

       Parameters
       ----------
       image : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       brightness_factor : tuple of 2 floats, optional
           Range of brightness' intensity. E.g. ``(0.1, 0.3)``.

       mode : str, optional
           One of ``2D`` or ``3D``.

       Returns
       -------
       image : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``brightness_factor=(0.1,0.3)``, ``mode='mix'``, ``invert=False`` and ``invert_p=0``
       may result in:

       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_bright.png    | .. figure:: ../../../img/bright.png         |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_bright2.png   | .. figure:: ../../../img/bright2.png        |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+

       The grid is painted for visualization purposes.
    """

    if brightness_factor[0] == 0 and brightness_factor[1] == 0: return image

    # Force mode if 2D
    if image.ndim == 3: mode == '3D'

    if mode == '2D':
        b_factor = np.random.uniform(brightness_factor[0], brightness_factor[1], image.shape[-1]*3)
        for z in range(image.shape[2]):
            image[:, :, z] += b_factor[z*3]
            image[:, :, z] = np.clip(image[:, :, z], 0, 1)
    else:
        b_factor = np.random.uniform(brightness_factor[0], brightness_factor[1])
        image += b_factor
        image = np.clip(image, 0, 1)

    return image

def contrast(image, contrast_factor=(0,0), mode='2D'):
    """Contrast augmentation.

       Parameters
       ----------
       image : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       contrast_factor : tuple of 2 floats, optional
           Range of contrast's intensity. E.g. ``(0.1, 0.3)``.

       mode : str, optional
           One of ``2D`` or ``3D``.

       Returns
       -------
       image : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``contrast_factor=(0.1,0.3)``, ``mode='mix'``, ``invert=False`` and ``invert_p=0``
       may result in:

       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_contrast.png  | .. figure:: ../../../img/contrast.png       |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_contrast2.png | .. figure:: ../../../img/contrast2.png      |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+

       The grid is painted for visualization purposes.
    """

    if contrast_factor[0] == 0 and contrast_factor[1] == 0: return image

    # Force mode if 2D
    if image.ndim == 3: mode == '3D'

    if mode == '2D':
        c_factor = np.random.uniform(contrast_factor[0], contrast_factor[1], image.shape[-1]*3)
        for z in range(image.shape[2]):
            image[:, :, z] *= 1 + c_factor[z*3]
            image[:, :, z] = np.clip(image[:, :, z], 0, 1)
    else:
        c_factor = np.random.uniform(contrast_factor[0], contrast_factor[1])
        image *= 1 + c_factor
        image = np.clip(image, 0, 1)

    return image


def brightness_em(image, brightness_factor=(0,0),  mode='2D', invert=False, invert_p=0):
    """Randomly adjust brightness, randomly invert the color space and apply gamma correction. 

       Implementation based on `PyTorch Connectomics' grayscale.py
       <https://github.com/zudi-lin/pytorch_connectomics/blob/master/connectomics/data/augmentation/grayscale.py>`_.

       Parameters
       ----------
       image : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       brightness_factor : tuple of 2 floats, optional
           Range of brightness' intensity. E.g. ``(0.1, 0.3)``.

       mode : str, optional
           One of ``2D`` or ``3D``.

       invert : bool, optional
           Whether to invert the images.

       invert_p : float, optional
           Probability of inverting the images.

       Returns
       -------
       image : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``brightness_factor=(0.1,0.3)``, ``mode='mix'``, ``invert=False`` and ``invert_p=0``
       may result in:

       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_bright.png    | .. figure:: ../../../img/bright.png         |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_bright2.png   | .. figure:: ../../../img/bright2.png        |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+

       The grid is painted for visualization purposes.
    """

    if brightness_factor[0] == 0 and brightness_factor[1] == 0: return image

    # Force mode if 2D
    if img.ndim == 3: mode == '3D'

    b_factor = random.uniform(brightness_factor[0], brightness_factor[1])
    if mode == '2D':
        ran = np.random.rand(image.shape[-1]*3)
        for z in range(image.shape[2]):
            img = image[:, :, z]
            image[:, :, z] += (ran[z*3+1] - 0.5)*b_factor
            image[:, :, z] = np.clip(image[:, :, z], 0, 1)
            image[:, :, z] **= 2.0**(ran[z*3+2]*2 - 1)
    else:
        ran = np.random.rand(3)
        image += (ran[1] - 0.5)*b_factor
        image = np.clip(image, 0, 1)
        image **= 2.0**(ran[2]*2 - 1)

    if invert and random.uniform(0, 1) < invert_p:
        image = 1.0-image
        image = np.clip(image, 0, 1)

    return image


def contrast_em(image, contrast_factor=(0,0), mode='2D', invert=False, invert_p=0):
    """Contrast augmentation. Randomly invert the color space and apply gamma correction. 

       Implementation based on `PyTorch Connectomics' grayscale.py
       <https://github.com/zudi-lin/pytorch_connectomics/blob/master/connectomics/data/augmentation/grayscale.py>`_.

       Parameters
       ----------
       image : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       contrast_factor : tuple of 2 floats, optional
           Range of contrast's intensity. E.g. ``(0.1, 0.3)``.

       mode : str, optional
           One of ``2D`` or ``3D``.

       invert : bool, optional
           Whether to invert the image.

       invert_p : float, optional
           Probability of inverting the image.

       Returns
       -------
       image : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``contrast_factor=(0.1,0.3)``, ``mode='mix'``, ``invert=False`` and ``invert_p=0``
       may result in:

       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_contrast.png  | .. figure:: ../../../img/contrast.png       |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_contrast2.png | .. figure:: ../../../img/contrast2.png      |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+

       The grid is painted for visualization purposes.
    """
    if contrast_factor[0] == 0 and contrast_factor[1] == 0: return image

    # Force mode if 2D
    if image.ndim == 3: mode == '3D'

    c_factor = random.uniform(contrast_factor[0], contrast_factor[1])
    if mode == '2D':
        ran = np.random.rand(image.shape[-1]*3)
        for z in range(image.shape[2]):
            img = image[:, :, z]
            image[:, :, z] *= 1 + (ran[z*3] - 0.5)*c_factor
            image[:, :, z] = np.clip(image[:, :, z], 0, 1)
            image[:, :, z] **= 2.0**(ran[z*3+2]*2 - 1)
    else:
        ran = np.random.rand(3)
        image *= 1 + (ran[0] - 0.5)*c_factor
        image = np.clip(image, 0, 1)
        image **= 2.0**(ran[2]*2 - 1)

    if invert and random.uniform(0, 1) < invert_p:
        image = 1.0-image
        image = np.clip(image, 0, 1)

    return image


def missing_sections(img, iterations=(30,40)):
    """Augment the image by creating a black line in a random position.

       Implementation based on `PyTorch Connectomics' missing_parts.py
       <https://github.com/zudi-lin/pytorch_connectomics/blob/master/connectomics/data/augmentation/missing_parts.py>`_.

       Parameters
       ----------
       img : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       iterations : tuple of 2 ints, optional
           Iterations to dilate the missing line with. E.g. ``(30, 40)``.

       Returns
       -------
       out : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with ``iterations=(30,40)`` may result in:

       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_missing.png   | .. figure:: ../../../img/missing.png        |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+
       | .. figure:: ../../../img/orig_missing2.png  | .. figure:: ../../../img/missing2.png       |
       |   :width: 80%                               |   :width: 80%                               |
       |   :align: center                            |   :align: center                            |
       |                                             |                                             |
       |   Input image                               |   Augmented image                           |
       +---------------------------------------------+---------------------------------------------+

       The grid is painted for visualization purposes.
    """

    it = np.random.randint(iterations[0], iterations[1])

    num_section = img.shape[-1]
    slice_shape = img.shape[:2]
    transforms = {}

    out = img.copy()

    i=0
    while i < num_section:
        if np.random.rand() < 0.5:
            transforms[i] = _prepare_deform_slice(slice_shape, it)
            i += 2 # at most one deformed image in any consecutive 3 images
        i += 1

    num_section = img.shape[-1]
    for i in range(num_section):
        if i in transforms.keys():
            line_mask = transforms[i]
            mean = img[...,i].mean()
            out[:,:,i][line_mask] = mean

    return out


def _prepare_deform_slice(slice_shape, iterations):
    """Auxiliary function for missing_sections. """
    shape = (slice_shape[0], slice_shape[1])
    # randomly choose fixed x or fixed y with p = 1/2
    fixed_x = np.random.rand() < 0.5
    if fixed_x:
        x0, y0 = 0, np.random.randint(1, shape[1] - 2)
        x1, y1 = shape[0] - 1, np.random.randint(1, shape[1] - 2)
    else:
        x0, y0 = np.random.randint(1, shape[0] - 2), 0
        x1, y1 = np.random.randint(1, shape[0] - 2), shape[1] - 1

    # generate the mask of the line that should be blacked out
    line_mask = np.zeros(shape, dtype='bool')
    rr, cc = line(x0, y0, x1, y1)
    line_mask[rr, cc] = 1

    # generate vectorfield pointing towards the line to compress the image
    # first we get the unit vector representing the line
    line_vector = np.array([x1 - x0, y1 - y0], dtype='float32')
    line_vector /= np.linalg.norm(line_vector)
    # next, we generate the normal to the line
    normal_vector = np.zeros_like(line_vector)
    normal_vector[0] = - line_vector[1]
    normal_vector[1] = line_vector[0]

    # find the 2 components where coordinates are bigger / smaller than the line
    # to apply normal vector in the correct direction
    components, n_components = label(np.logical_not(line_mask).view('uint8'))
    assert n_components == 2, "%i" % n_components

    # dilate the line mask
    line_mask = binary_dilation(line_mask, iterations=iterations)

    return line_mask


def shuffle_channels(img):
    """Augment the image by shuffling its channels.

       Parameters
       ----------
       img : 3D/4D Numpy array
           Image to transform. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       Returns
       -------
       out : 3D/4D Numpy array
           Transformed image. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       Examples
       --------

       +-----------------------------------------------+-----------------------------------------------+
       | .. figure:: ../../../img/orig_chshuffle.png   | .. figure:: ../../../img/chshuffle.png        |
       |   :width: 80%                                 |   :width: 80%                                 |
       |   :align: center                              |   :align: center                              |
       |                                               |                                               |
       |   Input image                                 |   Augmented image                             |
       +-----------------------------------------------+-----------------------------------------------+

       The grid is painted for visualization purposes.
    """

    if img.ndim != 3 and img.ndim != 4:
        raise ValueError("Image is supposed to be 3 or 4 dimensions but provided {} image shape instead".format(img.shape))

    new_channel_order = np.random.permutation(img.shape[-1])
    return img[...,new_channel_order]


def grayscale(img):
    """Augment the image by converting it into grayscale.

       Parameters
       ----------
       img : 3D/4D Numpy array
           Image to transform. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       Returns
       -------
       out : 3D/4D Numpy array
           Transformed image. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       Examples
       --------

       +-----------------------------------------------+-----------------------------------------------+
       | .. figure:: ../../../img/orig_grayscale.png   | .. figure:: ../../../img/grayscale.png        |
       |   :width: 80%                                 |   :width: 80%                                 |
       |   :align: center                              |   :align: center                              |
       |                                               |                                               |
       |   Input image                                 |   Augmented image                             |
       +-----------------------------------------------+-----------------------------------------------+

       The grid is painted for visualization purposes.
    """

    if img.shape[-1] != 3:
        raise ValueError("Image is supposed to have 3 channels (RGB). Provided {} image shape instead".format(img.shape))

    return np.tile(np.expand_dims(np.mean(img, -1), -1), 3)


def GridMask(img, channels, z_size, ratio=0.6, d_range=(30,60), rotate=1, invert=False):
    """GridMask data augmentation presented in `GridMask Data Augmentation <https://arxiv.org/abs/2001.04086v1>`_.
       Code adapted from `<https://github.com/dvlab-research/GridMask/blob/master/imagenet_grid/utils/grid.py>`_.

       Parameters
       ----------
       img : 3D Numpy array
           Image to transform. E.g. ``(y, x, channels)``.

       channels : int
           Size of channel dimension. Used for 3D images as the channels have been merged with the z axis.

       z_size : int
           Size of z dimension. Used for 3D images as the z axis has been merged with the channels. Set to -1 to when
           do not want to be applied.

       ratio : tuple of floats, optional
           Range to choose the size of the areas to create.

       d_range : tuple of ints, optional
           Range to choose the ``d`` value in the original paper.

       rotate : float, optional
           Rotation of the mask in GridMask. Needs to be between ``[0,1]`` where 1 is 360 degrees.

       invert : bool, optional
           Whether to invert the mask.

       Returns
       -------
       out : 3D Numpy array
           Transformed image. E.g. ``(y, x, channels)``.

       Examples
       --------

       Calling this function with the default settings may result in:

       +----------------------------------------------+----------------------------------------------+
       | .. figure:: ../../../img/orig_GridMask.png   | .. figure:: ../../../img/GridMask.png        |
       |   :width: 80%                                |   :width: 80%                                |
       |   :align: center                             |   :align: center                             |
       |                                              |                                              |
       |   Input image                                |   Augmented image                            |
       +----------------------------------------------+----------------------------------------------+

       The grid is painted for visualization purposes.
    """

    h,w,c = img.shape

    # 1.5 * h, 1.5 * w works fine with the squared images
    # But with rectangular input, the mask might not be able to recover back to the input image shape
    # A square mask with edge length equal to the diagnoal of the input image
    # will be able to cover all the image spot after the rotation. This is also the minimum square.
    hh = math.ceil((math.sqrt(h*h + w*w)))

    d = np.random.randint(d_range[0], d_range[1])

    l = math.ceil(d*ratio)

    mask = np.ones((hh, hh), np.float32)
    st_h = np.random.randint(d)
    st_w = np.random.randint(d)
    for i in range(-1, hh//d+1):
            s = d*i + st_h
            t = s+l
            s = max(min(s, hh), 0)
            t = max(min(t, hh), 0)
            mask[s:t,:] *= 0
    for i in range(-1, hh//d+1):
            s = d*i + st_w
            t = s+l
            s = max(min(s, hh), 0)
            t = max(min(t, hh), 0)
            mask[:,s:t] *= 0
    r = np.random.randint(rotate)
    mask = Image.fromarray(np.uint8(mask))
    mask = mask.rotate(r)
    mask = np.asarray(mask)
    mask = mask[(hh-h)//2:(hh-h)//2+h, (hh-w)//2:(hh-w)//2+w]

    if not invert: mask = 1-mask
    if z_size != -1:
        _z_size = np.random.randint(d_range[2], d_range[3])
        cz = np.random.randint(0, z_size-_z_size)
        img[...,cz*channels:(cz*channels)+(_z_size*channels)] *= np.stack((mask,)*(_z_size*channels), axis=-1)
        return img
    else:
        return img*np.stack((mask,)*img.shape[-1], axis=-1)


def random_crop_pair(image, mask, random_crop_size, val=False, draw_prob_map_points=False, img_prob=None, weight_map=None,
        scale=1):
    """Random crop for an image and its mask. No crop is done in those dimensions that ``random_crop_size`` is greater than
       the input image shape in those dimensions. For instance, if an input image is ``400x150`` and ``random_crop_size``
       is ``224x224`` the resulting image will be ``224x150``.

       Parameters
       ----------
       image : Numpy 3D array
           Image. E.g. ``(y, x, channels)``.

       mask : Numpy 3D array
           Image mask. E.g. ``(y, x, channels)``.

       random_crop_size : 2 int tuple
           Size of the crop. E.g. ``(height, width)``.

       val : bool, optional
           If the image provided is going to be used in the validation data. This forces to crop from the origin,
           e. g. ``(0, 0)`` point.

       draw_prob_map_points : bool, optional
           To return the pixel chosen to be the center of the crop.

       img_prob : Numpy 3D array, optional
           Probability of each pixel to be chosen as the center of the crop. E. .g. ``(y, x, channels)``.

       weight_map : bool, optional
           Weight map of the given image. E.g. ``(y, x, channels)``.

       scale : int, optional
           Scale factor the second image given.

       Returns
       -------
       img : 2D Numpy array
           Crop of the given image. E.g. ``(y, x)``.

       weight_map : 2D Numpy array, optional
           Crop of the given image's weigth map. E.g. ``(y, x)``.

       ox : int, optional
           X coordinate in the complete image of the chose central pixel to make the crop.

       oy : int, optional
           Y coordinate in the complete image of the chose central pixel to make the crop.

       x : int, optional
           X coordinate in the complete image where the crop starts.

       y : int, optional
           Y coordinate in the complete image where the crop starts.
    """

    if weight_map is not None:
        img, we = image
    else:
        img = image

    height, width = img.shape[0], img.shape[1]
    dy, dx = random_crop_size[0], random_crop_size[1]
    if val:
        y, x, oy, ox = 0, 0, 0, 0
    else:
        if img_prob is not None:
            prob = img_prob.ravel()

            # Generate the random coordinates based on the distribution
            choices = np.prod(img_prob.shape)
            index = np.random.choice(choices, size=1, p=prob)
            coordinates = np.unravel_index(index, dims=img_prob.shape)
            x = int(coordinates[1][0])
            y = int(coordinates[0][0])
            ox = int(coordinates[1][0])
            oy = int(coordinates[0][0])

            # Adjust the coordinates to be the origin of the crop and control to
            # not be out of the image
            if y < int(random_crop_size[0]/2):
                y = 0
            elif y > img.shape[0] - int(random_crop_size[0]/2):
                y = img.shape[0] - random_crop_size[0]
            else:
                y -= int(random_crop_size[0]/2)

            if x < int(random_crop_size[1]/2):
                x = 0
            elif x > img.shape[1] - int(random_crop_size[1]/2):
                x = img.shape[1] - random_crop_size[1]
            else:
                x -= int(random_crop_size[1]/2)
        else:
            oy, ox = 0, 0
            x = np.random.randint(0, width - dx + 1) if width - dx +1 > 0 else 0
            y = np.random.randint(0, height - dy + 1) if height - dy +1 > 0 else 0

    # Super-resolution check
    if scale != 1:
        img_out_shape = img[y:(y+dy), x:(x+dx)].shape
        mask_out_shape = mask[y*scale:(y+dy)*scale, x*scale:(x+dx)*scale].shape
        s = [img_out_shape[0]*scale, img_out_shape[1]*scale]
        if all(x!=y for x,y in zip(s,mask_out_shape)):
            raise ValueError("Images can not be cropped to a PATCH_SIZE of {}. Inputs: LR image shape={} "
                "and HR image shape={}. When cropping the output shapes are {} and {}, for LR and HR images respectively. "
                "Try to reduce DATA.PATCH_SIZE".format(random_crop_size, img.shape, mask.shape, img_out_shape, mask_out_shape))

    if draw_prob_map_points == True:
        return img[y:(y+dy), x:(x+dx)], mask[y*scale:(y+dy)*scale, x*scale:(x+dx)*scale], oy, ox, y, x
    else:
        if weight_map is not None:
            return img[y:(y+dy), x:(x+dx)], mask[y*scale:(y+dy)*scale, x*scale:(x+dx)*scale], weight_map[y:(y+dy), x:(x+dx)]
        else:
            return img[y:(y+dy), x:(x+dx)], mask[y*scale:(y+dy)*scale, x*scale:(x+dx)*scale]


def random_3D_crop_pair(image, mask, random_crop_size, val=False, img_prob=None, weight_map=None, draw_prob_map_points=False,
        scale=1):
    """Extracts a random 3D patch from the given image and mask. No crop is done in those dimensions that ``random_crop_size`` is 
       greater than the input image shape in those dimensions. For instance, if an input image is ``10x400x150`` and ``random_crop_size``
       is ``10x224x224`` the resulting image will be ``10x224x150``.

       Parameters
       ----------
       image : 4D Numpy array
           Data to extract the patch from. E.g. ``(z, y, x, channels)``.

       mask : 4D Numpy array
           Data mask to extract the patch from. E.g. ``(z, y, x, channels)``.

       random_crop_size : 3D int tuple
           Shape of the patches to create. E.g. ``(z, y, x)``.

       val : bool, optional
           If the image provided is going to be used in the validation data. This forces to crop from the origin, e.g.
           ``(0, 0)`` point.

       img_prob : Numpy 4D array, optional
           Probability of each pixel to be chosen as the center of the crop. E. g. ``(z, y, x, channels)``.

       weight_map : bool, optional
           Weight map of the given image. E.g. ``(y, x, channels)``.

       draw_prob_map_points : bool, optional
           To return the voxel chosen to be the center of the crop.

       scale : int, optional
           Scale factor the second image given.

       Returns
       -------
       img : 4D Numpy array
           Crop of the given image. E.g. ``(z, y, x, channels)``.

       weight_map : 4D Numpy array, optional
           Crop of the given image's weigth map. E.g. ``(z, y, x, channels)``.
    
       oz : int, optional
           Z coordinate in the complete image of the chose central pixel to
           make the crop.

       oy : int, optional
           Y coordinate in the complete image of the chose central pixel to
           make the crop.

       ox : int, optional
           X coordinate in the complete image of the chose central pixel to
           make the crop.

       z : int, optional
           Z coordinate in the complete image where the crop starts.

       y : int, optional
           Y coordinate in the complete image where the crop starts.

       x : int, optional
           X coordinate in the complete image where the crop starts.
    """

    if weight_map is not None:
        vol, we = image
    else:
        vol = image

    deep, cols, rows = vol.shape[0], vol.shape[1], vol.shape[2]
    dz, dy, dx = random_crop_size
    if val:
        x, y, z, ox, oy, oz = 0, 0, 0, 0, 0, 0
    else:
        if img_prob is not None:
            prob = img_prob.ravel()

            # Generate the random coordinates based on the distribution
            choices = np.prod(img_prob.shape)
            index = np.random.choice(choices, size=1, p=prob)
            coordinates = np.unravel_index(index, shape=img_prob.shape)
            x = int(coordinates[2])
            y = int(coordinates[1])
            z = int(coordinates[0])
            ox = int(coordinates[2])
            oy = int(coordinates[1])
            oz = int(coordinates[0])

            # Adjust the coordinates to be the origin of the crop and control to
            # not be out of the volume
            if z < int(random_crop_size[0]/2):
                z = 0
            elif z > vol.shape[0] - int(random_crop_size[0]/2):
                z = vol.shape[0] - random_crop_size[0]
            else:
                z -= int(random_crop_size[0]/2)

            if y < int(random_crop_size[1]/2):
                y = 0
            elif y > vol.shape[1] - int(random_crop_size[1]/2):
                y = vol.shape[1] - random_crop_size[1]
            else:
                y -= int(random_crop_size[1]/2)

            if x < int(random_crop_size[2]/2):
                x = 0
            elif x > vol.shape[2] - int(random_crop_size[2]/2):
                x = vol.shape[2] - random_crop_size[2]
            else:
                x -= int(random_crop_size[2]/2)
        else:
            ox = 0
            oy = 0
            oz = 0
            z = np.random.randint(0, deep - dz + 1) if deep - dz +1 > 0 else 0
            y = np.random.randint(0, cols - dy + 1) if cols - dy +1 > 0 else 0
            x = np.random.randint(0, rows - dx + 1) if rows - dx +1 > 0 else 0

     # Super-resolution check
    if scale != 1:
        img_out_shape = vol[z:(z+dz), y:(y+dy), x:(x+dx)].shape
        mask_out_shape = mask[z:(z+dz), y*scale:(y+dy)*scale, x*scale:(x+dx)*scale].shape
        s = [img_out_shape[0], img_out_shape[1]*scale, img_out_shape[2]*scale]
        if all(x!=y for x,y in zip(s,mask_out_shape)):
            raise ValueError("Images can not be cropped to a PATCH_SIZE of {}. Inputs: LR image shape={} "
                "and HR image shape={}. When cropping the output shapes are {} and {}, for LR and HR images respectively. "
                "Try to reduce DATA.PATCH_SIZE".format(random_crop_size, vol.shape, mask.shape, img_out_shape, mask_out_shape))

    if draw_prob_map_points:
        return vol[z:(z+dz), y:(y+dy), x:(x+dx)], mask[z:(z+dz), y*scale:(y+dy)*scale, x*scale:(x+dx)*scale],\
               oz, oy, ox, z, y, x
    else:
        if weight_map is not None:
            return vol[z:(z+dz), y:(y+dy), x:(x+dx)], mask[z:(z+dz), y*scale:(y+dy)*scale, x*scale:(x+dx)*scale],\
                   weight_map[z:(z+dz), y:(y+dy), x:(x+dx)]
        else:
            return vol[z:(z+dz), y:(y+dy), x:(x+dx)], mask[z:(z+dz), y:(y+dy), x:(x+dx)]


def random_crop_single(image, random_crop_size, val=False, draw_prob_map_points=False, weight_map=None):
    """Random crop for a single image. No crop is done in those dimensions that ``random_crop_size`` is greater than
       the input image shape in those dimensions. For instance, if an input image is ``400x150`` and ``random_crop_size``
       is ``224x224`` the resulting image will be ``224x150``.

       Parameters
       ----------
       image : Numpy 3D array
           Image. E.g. ``(y, x, channels)``.

       random_crop_size : 2 int tuple
           Size of the crop. E.g. ``(y, x)``.

       val : bool, optional
           If the image provided is going to be used in the validation data. This forces to crop from the origin,
           e. g. ``(0, 0)`` point.

       draw_prob_map_points : bool, optional
           To return the pixel chosen to be the center of the crop.

       weight_map : bool, optional
           Weight map of the given image. E.g. ``(y, x, channels)``.

       Returns
       -------
       img : 2D Numpy array
           Crop of the given image. E.g. ``(y, x)``.

       weight_map : 2D Numpy array, optional
           Crop of the given image's weigth map. E.g. ``(y, x)``.

       oy : int, optional
           Y coordinate in the complete image of the chose central pixel to make the crop.

       ox : int, optional
           X coordinate in the complete image of the chose central pixel to make the crop.

       y : int, optional
           Y coordinate in the complete image where the crop starts.

       y : int, optional
           X coordinate in the complete image where the crop starts.
    """

    if weight_map is not None:
        img, we = image
    else:
        img = image

    height, width  = img.shape[0], img.shape[1]
    dy, dx = random_crop_size
    if val:
        x, y, z, ox, oy, oz = 0, 0, 0, 0, 0, 0
    else:
        oy, ox = 0, 0
        x = np.random.randint(0, width - dx + 1) if width - dx +1 > 0 else 0
        y = np.random.randint(0, height - dy + 1) if height - dy + 1 > 0 else 0

    if draw_prob_map_points == True:
        return img[y:(y+dy), x:(x+dx)], ox, oy, x, y
    else:
        if weight_map is not None:
            return img[y:(y+dy), x:(x+dx)], weight_map[y:(y+dy), x:(x+dx)]
        else:
            return img[y:(y+dy), x:(x+dx)]


def random_3D_crop_single(image, random_crop_size, val=False, draw_prob_map_points=False, weight_map=None):
    """Random crop for a single image. No crop is done in those dimensions that ``random_crop_size`` is greater than
       the input image shape in those dimensions. For instance, if an input image is ``50x400x150`` and ``random_crop_size``
       is ``30x224x224`` the resulting image will be ``30x224x150``.

       Parameters
       ----------
       image : Numpy 3D array
           Image. E.g. ``(z, y, x, channels)``.

       random_crop_size : 2 int tuple
           Size of the crop. E.g. ``(z, y, x)``.

       val : bool, optional
           If the image provided is going to be used in the validation data. This forces to crop from the origin,
           e. g. ``(0, 0)`` point.

       draw_prob_map_points : bool, optional
           To return the pixel chosen to be the center of the crop.

       weight_map : bool, optional
           Weight map of the given image. E.g. ``(z, y, x, channels)``.

       Returns
       -------
       img : 2D Numpy array
           Crop of the given image. E.g. ``(z, y, x)``.

       weight_map : 2D Numpy array, optional
           Crop of the given image's weigth map. E.g. ``(z, y, x)``.

       ox : int, optional
           Z coordinate in the complete image of the chose central pixel to make the crop.

       oy : int, optional
           Y coordinate in the complete image of the chose central pixel to make the crop.

       ox : int, optional
           X coordinate in the complete image of the chose central pixel to make the crop.

       z : int, optional
           Z coordinate in the complete image where the crop starts.

       y : int, optional
           Y coordinate in the complete image where the crop starts.

       x : int, optional
           X coordinate in the complete image where the crop starts.
    """

    if weight_map is not None:
        img, we = image
    else:
        img = image

    deep, cols, rows = img.shape[0], img.shape[1], img.shape[2]
    dz, dy, dx = random_crop_size
    if val:
        x, y, z, ox, oy, oz = 0, 0, 0, 0, 0, 0
    else:
        if img_prob is not None:
            prob = img_prob.ravel()

            # Generate the random coordinates based on the distribution
            choices = np.prod(img_prob.shape)
            index = np.random.choice(choices, size=1, p=prob)
            coordinates = np.unravel_index(index, shape=img_prob.shape)
            x = int(coordinates[2])
            y = int(coordinates[1])
            z = int(coordinates[0])
            ox = int(coordinates[2])
            oy = int(coordinates[1])
            oz = int(coordinates[0])

            # Adjust the coordinates to be the origin of the crop and control to
            # not be out of the volume
            if z < int(random_crop_size[0]/2):
                z = 0
            elif z > img.shape[0] - int(random_crop_size[0]/2):
                z = img.shape[0] - random_crop_size[0]
            else:
                z -= int(random_crop_size[0]/2)

            if y < int(random_crop_size[1]/2):
                y = 0
            elif y > img.shape[1] - int(random_crop_size[1]/2):
                y = img.shape[1] - random_crop_size[1]
            else:
                y -= int(random_crop_size[1]/2)

            if x < int(random_crop_size[2]/2):
                x = 0
            elif x > img.shape[2] - int(random_crop_size[2]/2):
                x = img.shape[2] - random_crop_size[2]
            else:
                x -= int(random_crop_size[2]/2)
        else:
            ox = 0
            oy = 0
            oz = 0
            z = np.random.randint(0, deep - dz + 1) if deep - dz +1 > 0 else 0
            y = np.random.randint(0, cols - dy + 1) if cols - dy + 1 > 0 else 0
            x = np.random.randint(0, rows - dx + 1) if rows - dx + 1 > 0 else 0

    if draw_prob_map_points:
        return img[z:(z+dz), y:(y+dy), x:(x+dx)], oz, oy, ox, z, y, x
    else:
        if weight_map is not None:
            return img[z:(z+dz), y:(y+dy), x:(x+dx)], weight_map[z:(z+dz), y:(y+dy), x:(x+dx)]
        else:
            return img[z:(z+dz), y:(y+dy), x:(x+dx)]



def center_crop_single(img, crop_shape):
    """Extract the central patch from a single image.

       Parameters
       ----------
       img : 3D/4D array
           Image. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       crop_shape : 2/3 int tuple
           Size of the crop. E.g. ``(y, x)`` or ``(z, y, x)``.

       Returns
       -------
       img : 3D/4D Numpy array
           Center crop of the given image. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.
    """
    if img.ndim == 4:
        z,y,x,c = img.shape
        startz = max(z//2 - crop_shape[0]//2, 0)
        starty = max(y//2 - crop_shape[1]//2, 0)
        startx = max(x//2 - crop_shape[2]//2, 0)  
        return img[startz:startz+crop_shape[0], starty:starty+crop_shape[1], startx:startx+crop_shape[2]]
    else:
        y,x,c = img.shape
        starty = max(y//2 - crop_shape[0]//2, 0)
        startx = max(x//2 - crop_shape[1]//2, 0)   
        return img[starty:starty+crop_shape[0], startx:startx+crop_shape[1]]

def resize_img(img, shape):
    """Resizes input image to given shape.

       Parameters
       ----------
       img : 3D/4D Numpy array
           Data to extract the patch from. E.g. ``(y, x, channels)`` for ``2D`` or  ``(z, y, x, channels)`` for ``3D``.

       crop_shape : 2D/3D int tuple
           Shape of the patches to create. E.g.  ``(y, x)`` for ``2D`` ``(z, y, x)`` for ``3D``.

       Returns
       -------
       img : 3D/4D Numpy array
           Resized image. E.g. ``(y, x, channels)`` for ``2D`` or  ``(z, y, x, channels)`` for ``3D``.
    """

    return resize(img, shape, order=1, mode='reflect', clip=True, preserve_range=True, anti_aliasing=True) 


def rotation(img, mask=None, heat=None, angles=[], mode="reflect", mask_type='as_mask'):
    """
    Apply a rotation to input ``image`` and ``mask`` (if provided). 

    Parameters
    ----------
    img : 3D/4D Numpy array
        Image to rotate. E.g. ``(y, x, channels)`` for ``2D`` or  ``(z, y, x, channels)`` for ``3D``.

    mask : 3D/4D Numpy array, optional
        Mask to rotate. E.g. ``(y, x, channels)`` for ``2D`` or  ``(z, y, x, channels)`` for ``3D``.

    heat : 3D/4D Numpy array, optional
        Heatmap (float mask) to rotate. E.g. ``(y, x, channels)`` for ``2D`` or  ``(z, y, x, channels)`` for ``3D``.

    angles : List of ints, optional
        List of angles to choose the rotation to be made. E.g. [90,180,360].

    mode : str, optional
        How to fill up the new values created. Options: ``reflect``, ``grid-mirror``,
        ``constant``, ``grid-constant``, ``nearest``, ``mirror``, ``grid-wrap``, ``wrap``.

    mask_type : str, optional
        How ``mask`` is going to be treated. Options: ``as_mask``, ``as_image``. With ``as_mask`` 
        the interpolation order will be 0 (nearest).

    Returns
    -------
    img : 3D/4D Numpy array
        Rotated image. E.g. ``(y, x, channels)`` for ``2D`` or  ``(z, y, x, channels)`` for ``3D``.

    mask : 3D/4D Numpy array, optional
        Rotated mask. E.g. ``(y, x, channels)`` for ``2D`` or  ``(z, y, x, channels)`` for ``3D``.
    
    heat : 3D/4D Numpy array, optional
        Rotated heatmap. Returned if ``mask`` is provided. E.g. ``(y, x, channels)`` for ``2D`` or  
        ``(z, y, x, channels)`` for ``3D``.
    """

    if len(angles) == 0:
        angle = np.random.randint(0, 360)
    elif isinstance(angles, tuple):
        assert len(angles) == 2, "If a tuple is provided it must have length of 2"
        angle = np.random.randint(angles[0], angles[1])
    elif isinstance(angles, list):
        angle = angles[np.random.randint(0, len(angles)-1)]
    else:
        raise ValueError("Not a list/tuple provided in 'angles'")

    img = rotate(img, angle=angle, mode=mode, reshape=False)

    if mask is not None:
        mask_order = 0 if mask_type == 'as_mask' else 1
        mask = rotate(mask, angle=angle, order=mask_order, mode=mode, reshape=False)
    if heat is not None:
        heat = rotate(heat, angle=angle, mode=mode, reshape=False)
    if mask is None:
        return img
    else:
        return img, mask, heat
            


# Adapted from https://github.com/L0-zhang/DFCAN-pytorch

import torch
import torch.nn as nn
import torch.fft

def fftshift2d(img, size_psc=128):
    bs,ch, h, w = img.shape
    fs11 = img[:,:, h//2:, w//2:]
    fs12 = img[:,:, h//2:, :w//2]
    fs21 = img[:,:, :h//2, w//2:]
    fs22 = img[:,:, :h//2, :w//2]
    output = torch.cat([torch.cat([fs11, fs21], dim=2), torch.cat([fs12, fs22], dim=2)], dim=3)
    return output

class RCAB(nn.Module):
    def __init__(self, size_psc=128): #size_psc：crop_size input_shape：depth
        super().__init__()
        self.size_psc = size_psc
        self.conv_gelu1 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding="same"),
            nn.GELU()
        )
        self.conv_gelu2 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding="same"),
            nn.GELU()
        )
        self.conv_relu1 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding="same"),
            nn.ReLU()
        )
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.conv_relu2 = nn.Sequential(
            nn.Conv2d(64, 4, kernel_size=1, stride=1, padding=0),
            nn.ReLU()
        )
        self.conv_sigmoid = nn.Sequential(
            nn.Conv2d(4, 64, kernel_size=1, stride=1, padding=0),
            nn.Sigmoid()
        )

    def forward(self,x,gamma=0.8):
        x0=x.to(torch.float32)
        x  = self.conv_gelu1(x)
        x  = self.conv_gelu2(x)
        x1 = x.to(torch.float32)
        x  = torch.fft.fftn(x.to(torch.float32),dim=(2,3))
        x  = torch.pow(torch.abs(x)+1e-8, gamma) #abs
        x  = fftshift2d(x, self.size_psc)
        x  = self.conv_relu1(x)
        x  = self.avg_pool(x)
        x  = self.conv_relu2(x)
        x  = self.conv_sigmoid(x)
        x  = x1*x
        x  = x0+x
        return x

class ResGroup(nn.Module):
    def __init__(self, n_RCAB=4, size_psc=128): #size_psc：crop_size input_shape：depth
        super().__init__()
        RCABs=[]
        for _ in range(n_RCAB):
            RCABs.append(RCAB(size_psc))
        self.RCABs=nn.Sequential(*RCABs)

    def forward(self,x):
        x0=x
        x=self.RCABs(x)
        x=x0+x
        return x

class DFCAN(nn.Module):
    """
    Fourier channel attention network (DFCAN) for super-resolution.
     
    References: `Evaluation and development of deep neural networks for image super-resolution in optical 
    microscopy <https://www.nature.com/articles/s41592-020-01048-5>`_.
    """
    def __init__(self, ndim, input_shape, scale=2, n_ResGroup = 4, n_RCAB = 4): 
        super().__init__()
        self.ndim = ndim
        size_psc = input_shape[0]
        self.input=nn.Sequential(
            nn.Conv2d(input_shape[-1], 64, kernel_size=3, stride=1, padding="same"),
            nn.GELU()
        )
        ResGroups=[]
        for _ in range(n_ResGroup):
            ResGroups.append(ResGroup(n_RCAB=n_RCAB, size_psc=size_psc))
        self.RGs = nn.Sequential(*ResGroups)
        self.conv_gelu=nn.Sequential(
            nn.Conv2d(64, 64*(scale ** 2), kernel_size=3, stride=1, padding="same"),
            nn.GELU()
        )
        self.pixel_shuffle = nn.PixelShuffle(scale)
        self.conv_sigmoid = nn.Sequential(
            nn.Conv2d(64, input_shape[-1], kernel_size=3, stride=1, padding="same"),
            nn.Sigmoid()
        )

    def forward(self,x):
        x=self.input(x)
        x=self.RGs(x)
        x=self.conv_gelu(x)
        x=self.pixel_shuffle(x) #upsampling
        x=self.conv_sigmoid(x)
        return x



import torch 
import torch.nn as nn
import torch.nn.functional as F

class ConvBlock(nn.Module):
    def __init__(self, conv, in_size, out_size, k_size, act=None, batch_norm=None, dropout=0, se_block=False):
        """
        Convolutional block.

        Parameters
        ----------
        conv : Torch convolutional layer
            Convolutional layer to use in the residual block. 

        in_size : array of ints
            Input feature maps of the convolutional layers.

        out_size : str, optional
            Output feature maps of the convolutional layers.

        k_size : 3 int tuple
            Height, width and depth of the convolution window.

        act : str, optional
            Activation layer to use. 

        batch_norm : nn.BatchNorm Torch layer, optional
            Batch normalization layer to use. 

        drop_value : float, optional
            Dropout value to be fixed.
        
        se_block : boolean, optional
            Whether to add Squeeze-and-Excitation blocks or not. 

        """
        super(ConvBlock, self).__init__()
        block = []

        block.append(conv(in_size, out_size, kernel_size=k_size, padding="same"))
        if batch_norm is not None:
            block.append(batch_norm(out_size))
        if act is not None:
            block.append(get_activation(act))
        if dropout > 0:
            block.append(nn.Dropout(dropout))
        if se_block:
            block.append(SqExBlock(out_size, ndim=2 if conv == nn.Conv2d else 3))

        self.block = nn.Sequential(*block)
        
    def forward(self, x):
        out = self.block(x)
        return out

class DoubleConvBlock(nn.Module):
    def __init__(self, conv, in_size, out_size, k_size, act=None, batch_norm=None, dropout=0, se_block=False):
        """
        Convolutional block.

        Parameters
        ----------
        conv : Torch convolutional layer
            Convolutional layer to use in the residual block. 

        in_size : array of ints
            Input feature maps of the convolutional layers.

        out_size : str, optional
            Output feature maps of the convolutional layers.

        k_size : 3 int tuple
            Height, width and depth of the convolution window.

        act : str, optional
            Activation layer to use. 

        batch_norm : nn.BatchNorm Torch layer, optional
            Batch normalization layer to use. 

        drop_value : float, optional
            Dropout value to be fixed.
        
        se_block : boolean, optional
            Whether to add Squeeze-and-Excitation blocks or not. 

        """
        super(DoubleConvBlock, self).__init__()
        block = []
        block.append(ConvBlock(conv=conv, in_size=in_size, out_size=out_size, k_size=k_size, act=act, 
            batch_norm=batch_norm, dropout=dropout, se_block=se_block))
        block.append(ConvBlock(conv=conv, in_size=out_size, out_size=out_size, k_size=k_size, act=act, 
            batch_norm=batch_norm, dropout=dropout, se_block=se_block))
        self.block = nn.Sequential(*block)
        
    def forward(self, x):
        out = self.block(x)
        return out

class UpBlock(nn.Module):
    def __init__(self, ndim, convtranspose, in_size, out_size, z_down, up_mode, conv, k_size, 
        act=None, batch_norm=None, dropout=0, attention_gate=False, se_block=False):
        """
        Convolutional upsampling block.

        Parameters
        ----------
        ndim : Torch convolutional layer
            Number of dimensions of the input data. 

        convtranspose : Torch convolutional layer
            Transpose convolutional layer to use. Only used if ``up_mode`` is ``'convtranspose'``. 

        in_size : array of ints
            Input feature maps of the convolutional layers.

        out_size : str, optional
            Output feature maps of the convolutional layers.

        z_down : int, optional
            Downsampling used in z dimension. 

        up_mode : str, optional
            Upsampling mode between ``'convtranspose'`` and ``'upsampling'``, which refers respectively
            to make an upsampling by appliying a transpose convolution (nn.ConvTranspose) or 
            upsampling layer (nn.Upsample). 

        conv : Torch convolutional layer
            Convolutional layer to use in the residual block. 

        k_size : 3 int tuple
            Height, width and depth of the convolution window.

        act : str, optional
            Activation layer to use. 

        batch_norm : nn.BatchNorm Torch layer, optional
            Batch normalization layer to use. 

        drop_value : float, optional
            Dropout value to be fixed.

        se_block : boolean, optional
            Whether to add Squeeze-and-Excitation blocks or not. 
        """
        super(UpBlock, self).__init__()
        self.ndim = ndim
        block = []
        mpool = (z_down, 2, 2) if ndim == 3 else (2, 2)
        if up_mode == 'convtranspose':
            block.append(convtranspose(in_size, out_size, kernel_size=mpool, stride=mpool))
        elif up_mode == 'upsampling':
            block.append(nn.Upsample(mode='bilinear' if ndim==2 else 'trilinear', scale_factor=mpool))
            block.append(conv(in_size, out_size, kernel_size=1))
        if batch_norm is not None:
            block.append(batch_norm(out_size))
        if act is not None:
            block.append(get_activation(act))
        self.up = nn.Sequential(*block)

        if attention_gate:
            self.attention_gate = AttentionBlock(conv=conv, in_size=out_size, out_size=out_size//2, batch_norm=batch_norm)
        else:
            self.attention_gate = None
        self.conv_block = DoubleConvBlock(conv=conv, in_size=out_size*2, out_size=out_size, k_size=k_size, 
            act=act, batch_norm=batch_norm, dropout=dropout, se_block=se_block)

    def forward(self, x, bridge):
        up = self.up(x)
        if self.attention_gate is not None:
            attn = self.attention_gate(up, bridge)
            out = torch.cat([up, attn], 1)
        else:
            out = torch.cat([up, bridge], 1)
        out = self.conv_block(out)
        return out

class AttentionBlock(nn.Module):
    def __init__(self, conv, in_size, out_size, batch_norm=None):
        """
        Attention block.

        Reference: `Attention U-Net: Learning Where to Look for the Pancreas <https://arxiv.org/abs/1804.03999>`_.

        Parameters
        ----------
        conv : Torch convolutional layer
            Convolutional layer to use in the residual block. 

        in_size : array of ints
            Input feature maps of the convolutional layers.

        out_size : str, optional
            Output feature maps of the convolutional layers.

        batch_norm : bool, optional
            To use batch normalization.
        """
        super(AttentionBlock, self).__init__()
        w_g = []
        w_g.append(conv(in_size, out_size, kernel_size=1, stride=1, padding=0, bias=True))
        if batch_norm is not None:
            w_g.append(batch_norm(out_size)) 
        self.w_g = nn.Sequential(*w_g)
        
        w_x = []
        w_x.append(conv(in_size, out_size, kernel_size=1, stride=1, padding=0, bias=True))
        if batch_norm is not None:
            w_x.append(batch_norm(out_size))
        self.w_x = nn.Sequential(*w_x)

        psi = []
        psi.append(conv(out_size, 1, kernel_size=1, stride=1, padding=0, bias=True))
        if batch_norm is not None:
            psi.append(batch_norm(1))
        psi.append(nn.Sigmoid())
        self.psi = nn.Sequential(*psi)

        self.relu = nn.ReLU(inplace=True)
        
    def forward(self, g, x):
        g1 = self.w_g(g)
        x1 = self.w_x(x)
        psi = self.relu(g1+x1)
        psi = self.psi(psi)
        return psi*x

class SqExBlock(nn.Module):
    """
    Squeeze-and-Excitation block from `Squeeze and Excitation Networks <https://arxiv.org/abs/1709.01507>`_.

    Credits: https://github.com/moskomule/senet.pytorch/blob/master/senet/se_module.py#L4"
    """
    def __init__(self, c, r=16, ndim=2):
        super().__init__()
        self.ndim = ndim
        self.squeeze = nn.AdaptiveAvgPool2d(1) if ndim == 2 else nn.AdaptiveAvgPool3d(1)
        self.excitation = nn.Sequential(
            nn.Linear(c, c // r, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(c // r, c, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        bs = x.shape[0]
        c = x.shape[1]
        y = self.squeeze(x).view(bs, c)
        y = self.excitation(y)
        if self.ndim == 2:
            y = y.view(bs, c, 1, 1)
        else:
            y = y.view(bs, c, 1, 1, 1)
        return x * y.expand_as(x)

class ResConvBlock(nn.Module):
    def __init__(self, conv, in_size, out_size, k_size, act=None, batch_norm=None, dropout=0, skip_k_size=1,
        skip_batch_norm=None, first_block=False):
        """
        Residual block.

        Parameters
        ----------
        conv : Torch convolutional layer
            Convolutional layer to use in the residual block. 

        in_size : array of ints
            Input feature maps of the convolutional layers.

        out_size : str, optional
            Output feature maps of the convolutional layers.

        k_size : 3 int tuple
            Height, width and depth of the convolution window.

        act : str, optional
            Activation layer to use. 

        batch_norm : nn.BatchNorm Torch layer, optional
            Batch normalization layer to use. 

        drop_value : float, optional
            Dropout value to be fixed.

        skip_k_size : int, optional
            Kernel size for the skip connection convolution. Used in resunet++.
        
        skip_batch_norm : nn.BatchNorm Torch layer, optional
            Batch normalization layer to use in the skip connection. Used in resunet++.

        first_block : float, optional
            To advice the function that it is the first residual block of the network, which avoids Full Pre-Activation
            layers (more info of Full Pre-Activation in `Identity Mappings in Deep Residual Networks
            <https://arxiv.org/pdf/1603.05027.pdf>`_).
        """
        super(ResConvBlock, self).__init__()
        block = []

        if not first_block:
            if batch_norm is not None:
                block.append(batch_norm(in_size))
            if act is not None:
                block.append(get_activation(act))

        block.append(ConvBlock(conv=conv, in_size=in_size, out_size=out_size, k_size=k_size, 
            act=act, batch_norm=batch_norm, dropout=dropout))
        block.append(ConvBlock(conv=conv, in_size=out_size, out_size=out_size, k_size=k_size))

        self.block = nn.Sequential(*block)

        block = []
        block.append(conv(in_size, out_size, kernel_size=skip_k_size, padding='same'))
        if skip_batch_norm is not None:
            block.append(skip_batch_norm(out_size))
        self.shortcut = nn.Sequential(*block)
    def forward(self, x):
        out = self.block(x) + self.shortcut(x)
        return out

class ResUpBlock(nn.Module):
    def __init__(self, ndim, convtranspose, in_size, out_size, in_size_bridge, z_down, up_mode, conv, k_size, 
        act=None, batch_norm=None, skip_k_size=1, skip_batch_norm=None, dropout=0):
        """
        Residual upsampling block.

        Parameters
        ----------
        ndim : Torch convolutional layer
            Number of dimensions of the input data. 

        convtranspose : Torch convolutional layer
            Transpose convolutional layer to use. Only used if ``up_mode`` is ``'convtranspose'``. 

        in_size : array of ints
            Input feature maps of the convolutional layers.

        out_size : int, optional
            Output feature maps of the convolutional layers.

        in_size_bridge : int, optional
            Output feature maps of the skip connection input. 

        z_down : int, optional
            Downsampling used in z dimension. 

        up_mode : str, optional
            Upsampling mode between ``'convtranspose'`` and ``'upsampling'``, which refers respectively
            to make an upsampling by appliying a transpose convolution (nn.ConvTranspose) or 
            upsampling layer (nn.Upsample). 

        conv : Torch convolutional layer
            Convolutional layer to use in the residual block. 

        k_size : 3 int tuple
            Height, width and depth of the convolution window.

        act : str, optional
            Activation layer to use. 

        batch_norm : nn.BatchNorm Torch layer, optional
            Batch normalization layer to use. 

        skip_k_size : int, optional
            Kernel size for the skip connection convolution. Used in resunet++.
        
        skip_batch_norm : nn.BatchNorm Torch layer, optional
            Batch normalization layer to use in the skip connection. Used in resunet++.

        drop_value : float, optional
            Dropout value to be fixed.
        """
        super(ResUpBlock, self).__init__()
        self.ndim = ndim
        mpool = (z_down, 2, 2) if ndim == 3 else (2, 2)
        if up_mode == 'convtranspose':
            self.up = convtranspose(in_size, in_size, kernel_size=mpool, stride=mpool)
        elif up_mode == 'upsampling':
            self.up = nn.Upsample(mode='bilinear' if ndim==2 else 'trilinear', scale_factor=mpool)
            
        self.conv_block = ResConvBlock(conv=conv, in_size=in_size+in_size_bridge, out_size=out_size, 
            k_size=k_size, act=act, batch_norm=batch_norm, dropout=dropout, skip_k_size=skip_k_size,
            skip_batch_norm=skip_batch_norm)

    def forward(self, x, bridge):
        up = self.up(x)
        out = torch.cat([up, bridge], 1)
        out = self.conv_block(out)
        return out

def get_activation(activation: str = 'relu') -> nn.Module:
    """
    Get the specified activation layer.

    Parameters
    ----------
    activation : str, optional
        One of ``'relu'``, ``'tanh'``, ``'leaky_relu'``, ``'elu'``, ``'gelu'``,
        ``'silu'``, ``'sigmoid'``, ``'softmax'``,``'swish'``, 'efficient_swish'``, 
        ``'linear'`` and ``'none'``.
    """
    assert activation in ["relu", "tanh", "leaky_relu", "elu", "gelu", "silu", "sigmoid",
        "softmax", "linear", "none"], "Get unknown activation key {}".format(activation)
    activation_dict = {
        "relu": nn.ReLU(),
        'tanh': nn.Tanh(),
        "leaky_relu": nn.LeakyReLU(negative_slope=0.2),
        "elu": nn.ELU(alpha=1.0),
        "gelu": nn.GELU(),
        "silu": nn.SiLU(),
        'sigmoid': nn.Sigmoid(),
        'softmax': nn.Softmax(dim=1),
        "linear": nn.Identity(),
        "none": nn.Identity()
    }
    return activation_dict[activation]


class ResUNetPlusPlus_AttentionBlock(nn.Module):
    """ Adapted from `here <https://github.com/rishikksh20/ResUnet/blob/master/core/modules.py>`_.
    """
    def __init__(self, conv, maxpool, input_encoder, input_decoder, output_dim, z_down=2, batch_norm=False):
        super(ResUNetPlusPlus_AttentionBlock, self).__init__()

        block = []
        if batch_norm is not None:
            block.append(batch_norm(input_encoder))
        block += [
            nn.ReLU(),
            conv(input_encoder, output_dim, 3, padding=1),
            maxpool((2, 2)) if conv == nn.Conv2d else maxpool((z_down, 2, 2))
        ]
        self.conv_encoder = nn.Sequential(*block)

        block = []
        if batch_norm is not None:
            block.append(batch_norm(input_decoder))
        block += [
            nn.ReLU(),
            conv(input_decoder, output_dim, 3, padding=1)
        ]
        self.conv_decoder = nn.Sequential(*block)

        block = []
        if batch_norm is not None:
            block.append(batch_norm(output_dim))
        block += [
            nn.ReLU(),
            conv(output_dim, 1, 1)
        ]
        self.conv_attn = nn.Sequential(*block)

    def forward(self, x1, x2):
        out = self.conv_encoder(x1) + self.conv_decoder(x2)
        out = self.conv_attn(out)
        return out * x2

class ASPP(nn.Module):
    """ Adapted from `here <https://github.com/rishikksh20/ResUnet/blob/master/core/modules.py>`_.
    """
    def __init__(self, conv, in_dims, out_dims, batch_norm=None, rate=[6, 12, 18]):
        super(ASPP, self).__init__()

        block = [
            conv(
                in_dims, out_dims, 3, stride=1, padding=rate[0], dilation=rate[0]
            ),
            nn.ReLU(inplace=True)
        ]
        if batch_norm is not None:
            block.append(batch_norm(out_dims))
        self.aspp_block1 = nn.Sequential(*block)
        block = [
            conv(
                in_dims, out_dims, 3, stride=1, padding=rate[1], dilation=rate[1]
            ),
            nn.ReLU(inplace=True),
        ]
        if batch_norm is not None:
            block.append(batch_norm(out_dims))    
        self.aspp_block2 = nn.Sequential(*block)
        block = [
            conv(
                in_dims, out_dims, 3, stride=1, padding=rate[2], dilation=rate[2]
            ),
            nn.ReLU(inplace=True)
        ]
        if batch_norm is not None:
            block.append(batch_norm(out_dims))
        self.aspp_block3 = nn.Sequential(*block)

        self.output = conv(len(rate) * out_dims, out_dims, 1)

    def forward(self, x):
        x1 = self.aspp_block1(x)
        x2 = self.aspp_block2(x)
        x3 = self.aspp_block3(x)
        out = torch.cat([x1, x2, x3], dim=1)
        return self.output(out)


import os
import h5py
import numpy as np
from skimage.io import imread
from tqdm import tqdm
from skimage import measure, feature
from scipy import ndimage
from PIL import ImageEnhance, Image
from os import path

img_shape = (4096, 4096)
nput_dir = "/home/user/input"
output_dir = "/home/user/output"
h5_filename = '0_human_instance_seg_pred.h5'
#h5_filename = '1_rat_instance_seg_pred.h5'

input_ids = sorted(next(os.walk(input_dir))[2])
h5file_name = os.path.join(output_dir, h5_filename)
os.makedirs(output_dir, exist_ok=True)

# Allocate memory for the predictions
pred_stack = np.zeros((len(input_ids),) + img_shape, dtype=np.int64)

# Read all the images
for n, id_ in tqdm(enumerate(input_ids), total=len(input_ids)):
    img = imread(os.path.join(input_dir, id_))
    pred_stack[n] = img

# Create the h5 file (using lzf compression to save space)
h5f = h5py.File(h5file_name, 'w')
h5f.create_dataset('main', data=pred_stack, compression="lzf")
h5f.close()




import os
import time
import warnings
import numpy as np
import math
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from skimage.io import imsave, imread

import numpy as np
import torch

class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience.
       Copied from: https://github.com/Bjarten/early-stopping-pytorch/blob/master/pytorchtools.py
    """
    def __init__(self, patience=7, delta=0, trace_func=print):
        """
        Parameters
        ----------
        patience : int, optional
            How long to wait after last time validation loss improved.

        delta : float, optional
            Minimum change in the monitored quantity to qualify as an improvement.

        trace_func : function, optional
            Trace print function.
        """
        self.patience = patience
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        self.trace_func = trace_func

    def __call__(self, val_loss):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
        elif score < self.best_score + self.delta:
            self.counter += 1
            self.trace_func(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.val_loss_min = val_loss
            self.counter = 0

# class MAE_callback():
#     """ Save MAE process in selected epochs. """
#     def __init__(self, test_gen, epoch_interval=None, out_dir="MAE_checks"):
#         self.epoch_interval = epoch_interval
#         self.test_gen = test_gen
#         self.out_dir = out_dir

#     def on_epoch_end(self, epoch, logs=None):
#         if self.epoch_interval and epoch != 0 and epoch % self.epoch_interval == 0:
#             test_images = next(iter(self.test_gen))

#             idx = np.random.choice(test_images.shape[0])
#             cmap = 'gray' if test_images[idx].shape[-1] == 1 else ''
#             test_images_patches = self.model.patchify(test_images)[idx]

#             latent, mask, ids_restore, unmasked_ids, masked_ids = self.model.encoder(test_images, training=False)
#             reconstructed = self.model.decoder(latent, training=False, ids_restore=ids_restore)[idx] 
#             masked_img = self.model.generate_masked_image(test_images_patches, unmasked_ids[idx])
            
#             fig = plt.figure(figsize=(30, 5), constrained_layout=True)
#             grid = gridspec.GridSpec(1, 5, figure=fig)

#             ax1 = fig.add_subplot(grid[0])
#             ax1.imshow(test_images[idx], cmap=cmap)
#             ax1.set_title(f"Original: {epoch:03d}")

#             rows, cols = self.model.encoder.patch_embed.grid_size
#             patch_size = self.model.encoder.patch_embed.patch_size[0]
#             gs00 = gridspec.GridSpecFromSubplotSpec(rows, cols, subplot_spec=grid[1])
#             for i in range(rows):
#                 for j in range(cols):
#                     ax = fig.add_subplot(gs00[i, j], xticklabels=[],yticklabels=[])
#                     ax.imshow(tf.reshape(test_images_patches[(i*cols)+j],(patch_size, patch_size)), cmap=cmap)
#                     ax.set_xticks([])
#                     ax.set_yticks([])

#             ax3 = fig.add_subplot(grid[2])
#             ax3.imshow(self.model.unpatchify(test_images_patches), cmap=cmap)
#             ax3.set_title(f"Unpatchify: {epoch:03d}")

#             ax4 = fig.add_subplot(grid[3])
#             ax4.imshow(self.model.unpatchify(masked_img), cmap=cmap)
#             ax4.set_title(f"Masked: {epoch:03d}")

#             ax5 = fig.add_subplot(grid[4])
#             ax5.imshow(self.model.unpatchify(reconstructed), cmap=cmap)
#             ax5.set_title(f"Reconstructed: {epoch:03d}")
            
#             os.makedirs(self.out_dir, exist_ok=True)
#             f = os.path.join(self.out_dir, "img_{}.png".format(epoch))
#             plt.savefig(f)
#             plt.close()
#             print("Saving MAE callback image . . .")
           



import os
import torch
import math
import numpy as np
from tqdm import tqdm
from torchmetrics.image import PeakSignalNoiseRatio

from biapy.data.data_2D_manipulation import crop_data_with_overlap, merge_data_with_overlap
from biapy.data.data_3D_manipulation import crop_3D_data_with_overlap, merge_3D_data_with_overlap
from biapy.data.post_processing.post_processing import ensemble8_2d_predictions, ensemble16_3d_predictions
from biapy.utils.util import save_tif
from biapy.utils.misc import to_pytorch_format, to_numpy_format
from biapy.engine.base_workflow import Base_Workflow
from biapy.data.pre_processing import create_ssl_source_data_masks, denormalize, undo_norm_range01
from biapy.engine.metrics import MaskedAutoencoderViT_loss

class Self_supervised_Workflow(Base_Workflow):
    """
    Self supervised workflow where the goal is to pretrain the backbone model by solving a so-called 
    pretext task without labels. This way, the model learns a representation that can be later transferred 
    to solve a downstream task in a labeled (but smaller) dataset. More details in `our documentation 
    <https://biapy.readthedocs.io/en/latest/workflows/self_supervision.html>`_.  

    Parameters
    ----------
    cfg : YACS configuration
        Running configuration.
    
    Job_identifier : str
        Complete name of the running job.

    device : Torch device
        Device used. 

    args : argpase class
        Arguments used in BiaPy's call. 
    """
    def __init__(self, cfg, job_identifier, device, args, **kwargs):
        super(Self_supervised_Workflow, self).__init__(cfg, job_identifier, device, args, **kwargs)
        
        self.prepare_ssl_data()

        # From now on, no modification of the cfg will be allowed
        self.cfg.freeze()

        # Activations for each output channel:
        # channel number : 'activation'
        self.activations = [{':': 'Linear'}]

        # Workflow specific training variables
        self.mask_path = None
        if cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == 'masking':
            self.load_Y_val = False
        else:
            self.mask_path = cfg.DATA.TRAIN.GT_PATH
            self.load_Y_val = True

    def define_metrics(self):
        """
        Definition of self.metrics, self.metric_names and self.loss variables.
        """
        self.metrics = [PeakSignalNoiseRatio().to(self.device)]
        self.metric_names = ["PSNR"]
        if self.cfg.MODEL.ARCHITECTURE == 'mae':
            print("Overriding 'LOSS.TYPE' to set it to MSE loss (masking patches)")
            self.loss = self.MaskedAutoencoderViT_loss_wrapper
        else:
            print("Overriding 'LOSS.TYPE' to set it to L1 loss")
            self.loss = torch.nn.L1Loss()

    def MaskedAutoencoderViT_loss_wrapper(self, output, targets):
        """
        Unravel MAE loss.
        """
        # Targets not used because the loss has been already calculated
        loss, pred, mask = output
        return loss

    def metric_calculation(self, output, targets, metric_logger=None):
        """
        Execution of the metrics defined in :func:`~define_metrics` function. 

        Parameters
        ----------
        output : Torch Tensor
            Prediction of the model. 

        targets : Torch Tensor
            Ground truth to compare the prediction with. 

        metric_logger : MetricLogger, optional
            Class to be updated with the new metric(s) value(s) calculated. 
        
        Returns
        -------
        value : float
            Value of the metric for the given prediction. 
        """
        # Calculate PSNR
        if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == 'masking':
            _, pred, _ = output
            pred = self.model_without_ddp.unpatchify(pred)
        else:
            pred = output
        with torch.no_grad():
            train_psnr = self.metrics[0](pred, targets)
            train_psnr = train_psnr.item() if not torch.isnan(train_psnr) else 0
            if metric_logger is not None:
                metric_logger.meters[self.metric_names[0]].update(train_psnr)
            else:
                return train_psnr

    def prepare_targets(self, targets, batch):
        """
        Location to perform any necessary data transformations to ``targets``
        before calculating the loss.

        Parameters
        ----------
        targets : Torch Tensor
            Ground truth to compare the prediction with.

        batch : Torch Tensor
            Prediction of the model. 

        Returns
        -------
        targets : Torch tensor
            Resulting targets. 
        """
        if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == 'masking':
            # Swap with original images so we can calculate PSNR metric afterwards
            return batch.to(self.device)
        else:
            return to_pytorch_format(targets, self.axis_order, self.device, dtype=self.loss_dtype)

    def process_sample(self, norm): 
        """
        Function to process a sample in the inference phase. 

        Parameters
        ----------
        norm : List of dicts
            Normalization used during training. Required to denormalize the predictions of the model.
        """
        original_data_shape = self._X.shape
    
        # Crop if necessary
        if self._X.shape[1:-1] != self.cfg.DATA.PATCH_SIZE[:-1]:
            if self.cfg.PROBLEM.NDIM == '2D':
                self._X = crop_data_with_overlap(self._X, self.cfg.DATA.PATCH_SIZE, overlap=self.cfg.DATA.TEST.OVERLAP, 
                    padding=self.cfg.DATA.TEST.PADDING, verbose=self.cfg.TEST.VERBOSE)
            else:
                self._X = crop_3D_data_with_overlap(self._X[0], self.cfg.DATA.PATCH_SIZE, overlap=self.cfg.DATA.TEST.OVERLAP, 
                    padding=self.cfg.DATA.TEST.PADDING, verbose=self.cfg.TEST.VERBOSE, 
                    median_padding=self.cfg.DATA.TEST.MEDIAN_PADDING)

        # Predict each patch
        if self.cfg.TEST.AUGMENTATION:
            for k in tqdm(range(self._X.shape[0]), leave=False):
                if self.cfg.PROBLEM.NDIM == '2D':
                    p = ensemble8_2d_predictions(self._X[k], n_classes=self.cfg.MODEL.N_CLASSES,
                            pred_func=(
                                lambda img_batch_subdiv: 
                                    to_numpy_format(
                                        self.apply_model_activations(
                                            self.model(to_pytorch_format(img_batch_subdiv, self.axis_order, self.device)),
                                            ), 
                                        self.axis_order_back
                                    )
                            )
                        )
                else:
                    p = ensemble16_3d_predictions(self._X[k], batch_size_value=self.cfg.TRAIN.BATCH_SIZE,
                            pred_func=(
                                lambda img_batch_subdiv: 
                                    to_numpy_format(
                                        self.apply_model_activations(
                                            self.model(to_pytorch_format(img_batch_subdiv, self.axis_order, self.device)),
                                            ), 
                                        self.axis_order_back
                                    )
                            )
                        )
                if 'pred' not in locals():
                    pred = np.zeros((self._X.shape[0],)+p.shape, dtype=self.dtype)
                pred[k] = p
        else:
            l = int(math.ceil(self._X.shape[0]/self.cfg.TRAIN.BATCH_SIZE))
            for k in tqdm(range(l), leave=False):
                top = (k+1)*self.cfg.TRAIN.BATCH_SIZE if (k+1)*self.cfg.TRAIN.BATCH_SIZE < self._X.shape[0] else self._X.shape[0]                
                with torch.cuda.amp.autocast():
                    p = self.model(to_pytorch_format(self._X[k*self.cfg.TRAIN.BATCH_SIZE:top], self.axis_order, self.device))
                    if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == "masking":
                        loss, p, mask = p
                        p = self.apply_model_activations(p)
                        p, m, pv = self.model.save_images(to_pytorch_format(self._X[k*self.cfg.TRAIN.BATCH_SIZE:top], self.axis_order, self.device), 
                            p, mask, self.dtype)
                    else:
                        p = self.apply_model_activations(p)
                        p = to_numpy_format(p, self.axis_order_back)
                    
                if 'pred' not in locals():
                    pred = np.zeros((self._X.shape[0],)+p.shape[1:], dtype=self.dtype)
                    if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == "masking":
                        pred_mask = np.zeros((self._X.shape[0],)+p.shape[1:], dtype=self.dtype)
                        pred_visi = np.zeros((self._X.shape[0],)+p.shape[1:], dtype=self.dtype)
                pred[k*self.cfg.TRAIN.BATCH_SIZE:top] = p
                if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == "masking":
                    pred_mask[k*self.cfg.TRAIN.BATCH_SIZE:top] = m
                    pred_visi[k*self.cfg.TRAIN.BATCH_SIZE:top] = pv

        # Delete self._X as in 3D there is no full image
        if self.cfg.PROBLEM.NDIM == '3D':
            del self._X, p

        # Reconstruct the predictions
        if original_data_shape[1:-1] != self.cfg.DATA.PATCH_SIZE[:-1]:
            if self.cfg.PROBLEM.NDIM == '3D': original_data_shape = original_data_shape[1:]
            f_name = merge_data_with_overlap if self.cfg.PROBLEM.NDIM == '2D' else merge_3D_data_with_overlap
            pred = f_name(pred, original_data_shape[:-1]+(pred.shape[-1],), padding=self.cfg.DATA.TEST.PADDING, 
                overlap=self.cfg.DATA.TEST.OVERLAP, verbose=self.cfg.TEST.VERBOSE)
            if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == "masking":
                pred_mask = f_name(pred_mask, original_data_shape[:-1]+(pred_mask.shape[-1],), padding=self.cfg.DATA.TEST.PADDING, 
                    overlap=self.cfg.DATA.TEST.OVERLAP, verbose=self.cfg.TEST.VERBOSE)
                pred_visi = f_name(pred_visi, original_data_shape[:-1]+(pred_visi.shape[-1],), padding=self.cfg.DATA.TEST.PADDING, 
                    overlap=self.cfg.DATA.TEST.OVERLAP, verbose=self.cfg.TEST.VERBOSE)
        else:
            pred = pred[0]
            if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == "masking":
                pred_mask = pred_mask[0]
                pred_visi = pred_visi[0]

        # Undo normalization
        x_norm = norm[0]
        if x_norm['type'] == 'div':
            pred = undo_norm_range01(pred, x_norm)
        else:
            pred = denormalize(pred, x_norm['mean'], x_norm['std'])  
            
            if x_norm['orig_dtype'] not in [np.dtype('float64'), np.dtype('float32'), np.dtype('float16')]:
                pred = np.round(pred)
                minpred = np.min(pred)                                                                                                
                pred = pred+abs(minpred)

            pred = pred.astype(x_norm['orig_dtype'])
            
        # Save image
        if self.cfg.PATHS.RESULT_DIR.PER_IMAGE != "":
            fname, fext = os.path.splitext(self.processing_filenames[0])
            save_tif(np.expand_dims(pred,0), self.cfg.PATHS.RESULT_DIR.PER_IMAGE, self.processing_filenames, verbose=self.cfg.TEST.VERBOSE)
            if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == "masking":
                save_tif(np.expand_dims(pred_mask,0), self.cfg.PATHS.RESULT_DIR.PER_IMAGE, [fname+"_masked.tif"], verbose=self.cfg.TEST.VERBOSE)
                save_tif(np.expand_dims(pred_visi,0), self.cfg.PATHS.RESULT_DIR.PER_IMAGE, [fname+"_reconstruction_and_visible.tif"], verbose=self.cfg.TEST.VERBOSE)    

    def torchvision_model_call(self, in_img, is_train=False):
        """
        Call a regular Pytorch model.

        Parameters
        ----------
        in_img : Tensor
            Input image to pass through the model.

        is_train : bool, optional
            Whether if the call is during training or inference. 

        Returns
        -------
        prediction : Tensor 
            Image prediction. 
        """
        pass

    def after_merge_patches(self, pred):
        """
        Steps need to be done after merging all predicted patches into the original image.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        pass

    def after_merge_patches_by_chunks_proccess_patch(self, filename):
        """
        Place any code that needs to be done after merging all predicted patches into the original image
        but in the process made chunk by chunk. This function will operate patch by patch defined by 
        ``DATA.PATCH_SIZE``.

        Parameters
        ----------
        filename : List of str
            Filename of the predicted image H5/Zarr.  
        """
        pass

    def after_full_image(self, pred):
        """
        Steps that must be executed after generating the prediction by supplying the entire image to the model.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        pass

    def after_all_images(self):
        """
        Steps that must be done after predicting all images. 
        """
        pass
    
    def normalize_stats(self, image_counter):
        """
        Normalize statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to average the metrics.
        """
        pass

    def print_stats(self, image_counter):
        """
        Print statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to call ``normalize_stats``.
        """
        self.normalize_stats(image_counter)


    def prepare_ssl_data(self):
        """
        Creates self supervised "ground truth" images, if ``crappify`` was selected, to train the model based 
        on the input images provided. They will be saved in a separate folder in the root path of the inout images. 
        """
        if self.cfg.PROBLEM.SELF_SUPERVISED.PRETEXT_TASK == "masking":
            print("No SSL data needs to be prepared for masking, as it will be generated on the fly")
            return

        print("############################")
        print("#  PREPARE DETECTION DATA  #")
        print("############################")

        # Create selected channels for train data
        if self.cfg.TRAIN.ENABLE:
            create_mask = False
            if not os.path.isdir(self.cfg.DATA.TRAIN.SSL_SOURCE_DIR):
                print("You select to create detection masks from given .csv files but no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TRAIN.SSL_SOURCE_DIR' "
                    "path, this process will be done just once!".format(self.cfg.DATA.TRAIN.SSL_SOURCE_DIR))
                create_mask = True
            else:
                if len(next(os.walk(self.cfg.DATA.TRAIN.SSL_SOURCE_DIR))[2]) != len(next(os.walk(self.cfg.DATA.TRAIN.PATH))[2]):
                    print("Different number of files found in {} and {}. Trying to create the the rest again"
                        .format(self.cfg.DATA.TRAIN.GT_PATH, self.cfg.DATA.TRAIN.SSL_SOURCE_DIR))
                    create_mask = True 
                else:
                    print("Train source data found in {}".format(self.cfg.DATA.TRAIN.SSL_SOURCE_DIR))   
            if create_mask:
                create_ssl_source_data_masks(self.cfg, data_type='train')

        # Create selected channels for val data
        if self.cfg.TRAIN.ENABLE and not self.cfg.DATA.VAL.FROM_TRAIN:
            create_mask = False
            if not os.path.isdir(self.cfg.DATA.VAL.SSL_SOURCE_DIR):
                print("You select to create detection masks from given .csv files but no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.VAL.SSL_SOURCE_DIR' "
                    "path, this process will be done just once!".format(self.cfg.DATA.VAL.SSL_SOURCE_DIR))
                create_mask = True
            else:
                if len(next(os.walk(self.cfg.DATA.VAL.SSL_SOURCE_DIR))[2]) != len(next(os.walk(self.cfg.DATA.VAL.PATH))[2]):
                    print("Different number of files found in {} and {}. Trying to create the the rest again"
                        .format(self.cfg.DATA.VAL.GT_PATH, self.cfg.DATA.VAL.SSL_SOURCE_DIR))
                    create_mask = True   
                else:
                    print("Validation source data found in {}".format(self.cfg.DATA.VAL.SSL_SOURCE_DIR)) 
            if create_mask:         
                create_ssl_source_data_masks(self.cfg, data_type='val')

        # Create selected channels for test data
        if self.cfg.TEST.ENABLE:
            create_mask = False
            if not os.path.isdir(self.cfg.DATA.TEST.SSL_SOURCE_DIR):
                print("You select to create detection masks from given .csv files but no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TEST.SSL_SOURCE_DIR' "
                    "path, this process will be done just once!".format(self.cfg.DATA.TEST.SSL_SOURCE_DIR))
                create_mask = True
            else:
                if len(next(os.walk(self.cfg.DATA.TEST.SSL_SOURCE_DIR))[2]) != len(next(os.walk(self.cfg.DATA.TEST.PATH))[2]):
                    print("Different number of files found in {} and {}. Trying to create the the rest again"
                        .format(self.cfg.DATA.TEST.GT_PATH, self.cfg.DATA.TEST.SSL_SOURCE_DIR))
                    create_mask = True    
                else:
                    print("Test source data found in {}".format(self.cfg.DATA.TEST.SSL_SOURCE_DIR))
            if create_mask:
                create_ssl_source_data_masks(self.cfg, data_type='test')

        opts = []
        if self.cfg.TRAIN.ENABLE:
            print("DATA.TRAIN.PATH changed from {} to {}".format(self.cfg.DATA.TRAIN.PATH, self.cfg.DATA.TRAIN.SSL_SOURCE_DIR))
            print("DATA.TRAIN.GT_PATH changed from {} to {}".format(self.cfg.DATA.TRAIN.GT_PATH, self.cfg.DATA.TRAIN.PATH))
            opts.extend(['DATA.TRAIN.PATH', self.cfg.DATA.TRAIN.SSL_SOURCE_DIR, 'DATA.TRAIN.GT_PATH', self.cfg.DATA.TRAIN.PATH])
            if not self.cfg.DATA.VAL.FROM_TRAIN:
                print("DATA.VAL.PATH changed from {} to {}".format(self.cfg.DATA.VAL.PATH, self.cfg.DATA.VAL.SSL_SOURCE_DIR))
                print("DATA.VAL.GT_PATH changed from {} to {}".format(self.cfg.DATA.VAL.GT_PATH, self.cfg.DATA.VAL.PATH))
                opts.extend(['DATA.VAL.PATH', self.cfg.DATA.VAL.SSL_SOURCE_DIR, 'DATA.VAL.GT_PATH', self.cfg.DATA.VAL.PATH])
        if self.cfg.TEST.ENABLE:
            print("DATA.TEST.PATH changed from {} to {}".format(self.cfg.DATA.TEST.PATH, self.cfg.DATA.TEST.SSL_SOURCE_DIR))
            print("DATA.TEST.GT_PATH changed from {} to {}".format(self.cfg.DATA.TEST.GT_PATH, self.cfg.DATA.TEST.PATH))
            opts.extend(['DATA.TEST.PATH', self.cfg.DATA.TEST.SSL_SOURCE_DIR, 'DATA.TEST.GT_PATH', self.cfg.DATA.TEST.PATH]) 
        self.cfg.merge_from_list(opts)



import os
import sys
import numpy as np
from skimage.io import imread
from tqdm import tqdm

code_dir = "/home/user/BiaPy"
input_dir = "/home/user/input"
out_dir = "/home/user/out"
output_prefix = ""
crop_shape = (256, 256, 1)
overlap = (0, 0)
padding = (0, 0)
verbose = True

input_dir_x = os.path.join(input_dir, 'x')
input_dir_y = os.path.join(input_dir, 'y')

sys.path.insert(0, code_dir)
from biapy.data.data_2D_manipulation import crop_data_with_overlap
from biapy.utils.util import save_tif_pair_discard

ids = sorted(next(os.walk(input_dir_x))[2])
for n, id_ in tqdm(enumerate(ids), total=len(ids)):
    if id_.endswith('.npy'):
        img = np.load(os.path.join(input_dir_x, id_))
        mask = np.load(os.path.join(input_dir_y, id_))
    else:
        img = imread(os.path.join(input_dir_x, id_))
        mask = imread(os.path.join(input_dir_y, id_))
    img = np.squeeze(img)
    mask = np.squeeze(mask)

    if len(img.shape) == 2:
        img = np.expand_dims(img, axis=-1)
    if len(mask.shape) == 2:
        mask = np.expand_dims(mask, axis=-1)

    if len(img.shape) == 3:
        img = np.expand_dims(img, axis=0)
    if len(mask.shape) == 3:
        mask = np.expand_dims(mask, axis=0)

    if img.shape != crop_shape[:2]+(img.shape[-1],):
        img, mask = crop_data_with_overlap(img, crop_shape[:2]+(img.shape[-1],), data_mask=mask, overlap=overlap,
                                              padding=padding, verbose=verbose)
    else:
        img = np.expand_dims(img, axis=0)
        mask = np.expand_dims(mask, axis=0)

    filenames = []
    d = len(str(len(img)))
    for k in range(img.shape[0]):
        filenames.append(os.path.splitext(id_)[0]+"_crop"+str(k).zfill(d)+'.tif')
    save_tif_pair_discard(img, mask, out_dir, output_prefix, filenames)

print("Finished!")





import numpy as np
import os
from PIL import Image

from biapy.utils.util import save_tif
from biapy.data.generators.pair_base_data_generator import PairBaseDataGenerator

class Pair2DImageDataGenerator(PairBaseDataGenerator):
    """Custom 2D data generator based on `imgaug <https://github.com/aleju/imgaug-doc>`_
       and our own `augmentors.py <https://github.com/BiaPyX/BiaPy/blob/master/biapy/data/generators/augmentors.py>`_
       transformations. This generator will yield an image and its corresponding mask.

       Based on `microDL <https://github.com/czbiohub/microDL>`_ and
       `Shervine's blog <https://stanford.edu/~shervine/blog/keras-how-to-generate-data-on-the-fly>`_.
    """
    def __init__(self, **kwars):
        super().__init__(**kwars)

    def ensure_shape(self, img, mask):
        # Shape adjustment
        if img.ndim == 2:
            img = np.expand_dims(img, -1)
        else:
            if img.shape[0] <= 3: img = img.transpose((1,2,0)) 
        if self.Y_provided:    
            if mask.ndim == 2: 
                mask = np.expand_dims(mask, -1)
            else:
                if mask.shape[0] <= 3: mask = mask.transpose((1,2,0))

        # Super-resolution check. if random_crops_in_DA is activated the images have not been cropped yet,
        # so this check can not be done and it will be done in the random crop
        if not self.random_crops_in_DA and self.Y_provided and self.random_crop_scale != 1:
            s = [img.shape[0]*self.random_crop_scale, img.shape[1]*self.random_crop_scale]
            if all(x!=y for x,y in zip(s,mask.shape[:-1])):
                raise ValueError("Images loaded need to be LR and its HR version. LR shape:"
                    " {} vs HR shape {} is not x{} larger".format(img.shape[:-1], mask.shape[:-1], self.random_crop_scale))

        if self.convert_to_rgb and self.shape[-1] == 3 and img.shape[-1] == 1:
            img = np.repeat(img, 3, axis=-1)

        if self.Y_provided:
            return img, mask
        else: 
            return img

    def save_aug_samples(self, img, mask, orig_images, i, pos, out_dir, point_dict):    
        aux = np.expand_dims(orig_images['o_x'], 0).astype(np.float32)
        save_tif(aux, out_dir, [str(i)+"_"+str(pos)+'_orig_x'+self.trans_made+".tif"], verbose=False)

        aux = np.expand_dims(orig_images['o_y'], 0).astype(np.float32)
        save_tif(aux, out_dir, [str(i)+"_"+str(pos)+'_orig_y'+self.trans_made+".tif"], verbose=False)
 
        # Save transformed images/masks
        aux = np.expand_dims(img, 0).astype(np.float32)
        save_tif(aux, out_dir, [str(i)+"_"+str(pos)+'_x'+self.trans_made+".tif"], verbose=False)
        aux = np.expand_dims(mask, 0).astype(np.float32)
        save_tif(aux, out_dir, [str(i)+"_"+str(pos)+'_y'+self.trans_made+".tif"], verbose=False)

        # Save the original images with a point that represents the selected coordinates to be the center of
        # the crop
        if self.random_crops_in_DA and self.prob_map is not None:
            img, mask = self.load_sample(pos)

            img, mask = (img).astype(np.uint8), mask.astype(np.uint8)

            if self.shape[-1] == 1:
                im = Image.fromarray(np.repeat(img, 3, axis=2), 'RGB')
            else:
                im = Image.fromarray(img, 'RGB')
            px = im.load()

            # Paint the selected point in red
            p_size=6
            for col in range(point_dict['oy']-p_size, point_dict['oy']+p_size):
                for row in range(point_dict['ox']-p_size, point_dict['ox']+p_size):
                    if col >= 0 and col < img.shape[0] and \
                        row >= 0 and row < img.shape[1]:
                        px[row, col] = (255, 0, 0)

            # Paint a blue square that represents the crop made
            for row in range(point_dict['s_x'], point_dict['s_x']+self.shape[0]):
                px[row, point_dict['s_y']] = (0, 0, 255)
                px[row, point_dict['s_y']+self.shape[0]-1] = (0, 0, 255)
            for col in range(point_dict['s_y'], point_dict['s_y']+self.shape[0]):
                px[point_dict['s_x'], col] = (0, 0, 255)
                px[point_dict['s_x']+self.shape[0]-1, col] = (0, 0, 255)

            im.save(os.path.join(out_dir, str(i)+"_"+str(pos)+'_mark_x'+self.trans_made+'.png'))

            if mask.shape[-1] == 1:
                m = Image.fromarray(np.repeat(mask, 3, axis=2), 'RGB')
            else:
                m = Image.fromarray(mask, 'RGB')
            px = m.load()

            # Paint the selected point in red
            for col in range(point_dict['oy']-p_size, point_dict['oy']+p_size):
                for row in range(point_dict['ox']-p_size, point_dict['ox']+p_size):
                    if col >= 0 and col < mask.shape[0] and \
                        row >= 0 and row < mask.shape[1]:
                        px[row, col] = (255, 0, 0)

            # Paint a blue square that represents the crop made
            for row in range(point_dict['s_x'], point_dict['s_x']+self.shape[0]):
                px[row, point_dict['s_y']] = (0, 0, 255)
                px[row, point_dict['s_y']+self.shape[0]-1] = (0, 0, 255)
            for col in range(point_dict['s_y'], point_dict['s_y']+self.shape[0]):
                px[point_dict['s_x'], col] = (0, 0, 255)
                px[point_dict['s_x']+self.shape[0]-1, col] = (0, 0, 255)

            m.save(os.path.join(out_dir, str(i)+"_"+str(pos)+'_mark_y'+self.trans_made+'.png'))



from scipy.spatial import cKDTree
import argparse
import pandas as pd


def remove_close_points( point_list, radius ):
    """Remove all points from ``point_list`` that are at a ``radius``
       or less distance from each other.

       Parameters
       ----------
       point_list : list
           List of 3D points. E.g. ``((0,0,0), (1,1,1)``.

       radius : float
           Radius from each point to decide what points to keep. E.g. ``10``.

       Returns
       -------
       value : list
           New list of points after removing those at a distance of ``radius``
           or less from each other.
    """
    mynumbers = [tuple(point) for point in point_list] 

    tree = cKDTree(mynumbers) # build k-dimensional tree
    
    pairs = tree.query_pairs( radius ) # find all pairs closer than radius
    
    neighbors = {} # create dictionary of neighbors

    for i,j in pairs: # iterate over all pairs
        if i not in neighbors:
            neighbors[i] = {j}
        else:
            neighbors[i].add(j)
        if j not in neighbors:
            neighbors[j] = {i}
        else:
            neighbors[j].add(i)
            
    positions = [i for i in range(0, len( point_list ))]

    keep = []
    discard = set()
    for node in positions:
        if node not in discard: # if node already in discard set: skip
            keep.append(node) # add node to keep list
            discard.update(neighbors.get(node,set())) # add node's neighbors to discard set

    # points to keep
    new_point_list = [ point_list[i] for i in keep]
    
    return new_point_list
    
parser = argparse.ArgumentParser(description="Filter points from napari CSV to remove those closer than a given radius",
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)


parser.add_argument("input_csv", help="napari CSV file with input points")
parser.add_argument("output", help="output CSV file name")
parser.add_argument("radius", type=float, help="distance between points to be considered the same")

parser.add_argument("-res0", "--resolution_axis0", type=float, default=1.0, help="Voxel size in physical units in axis 0")
parser.add_argument("-res1", "--resolution_axis1", type=float, default=1.0, help="Voxel size in physical units in axis 1")
parser.add_argument("-res2", "--resolution_axis2", type=float, default=1.0, help="Voxel size in physical units in axis 2")
parser.add_argument("--verbose", help="increase output verbosity",
                    action="store_true")

args = vars(parser.parse_args())

input_csv = args['input_csv']
out_csv = args['output']
radius = args['radius']

z_res = args['resolution_axis0']
y_res = args['resolution_axis1']
x_res = args['resolution_axis2']

verbose = args['verbose']

# read input CSV
df = pd.read_csv( input_csv, index_col=False, usecols=['axis-0','axis-1','axis-2'] )

# apply calibration
df['axis-0'] = df['axis-0'].apply(lambda z: z * z_res)
df['axis-1'] = df['axis-1'].apply(lambda y: y * y_res)
df['axis-2'] = df['axis-2'].apply(lambda x: x * x_res)

# extract points as list
points = df.values.tolist()

if verbose:
    print( 'Initial number of points: ' + str( len( points ) ) )

# filter points
new_points = remove_close_points( points, radius )

if verbose:
    print( 'Final number of points: ' + str( len( new_points ) ) )

# remove calibration
z_axis = []
y_axis = []
x_axis = []
for i in range(0, len(new_points)):
    z_axis.append( round( new_points[i][0]/z_res ) )
    y_axis.append( round( new_points[i][1]/y_res ) )
    x_axis.append( round( new_points[i][2]/x_res ) )

# create data frame in Napari CSV format
df_save = pd.DataFrame(list(zip(z_axis,y_axis,x_axis)), columns =['axis-0', 'axis-1', 'axis-2'])
# save to file
df_save.to_csv( out_csv, index=True, index_label='index' )

if verbose:
    print( 'Filtered points saved in ' + out_csv )

